import numpy
from .matrix_operations import calc_m1_m2_inv_m1, \
    calc_m_q_inv_m, calc_m_v, calc_vector_product_v1_v2_v1,\
    calc_mm_as_m_q_inv_m, \
    calc_mm_as_m1_m2_inv_m1, \
    calc_m_v, calc_m1_m2_inv_m1
from .symmetry_elements import calc_multiplicity_by_atom_symm_elems, calc_symm_flags
from .unit_cell import calc_eq_ccs_by_unit_cell_parameters, calc_m_m_by_unit_cell_parameters,\
    calc_volume_uc_by_unit_cell_parameters, \
    calc_sthovl_by_unit_cell_parameters
from .local_susceptibility import calc_chi_direct_norm_with_symmetry, calc_m_r_inv_m

from .symmetry_constraints import calc_sc_chi_full

from .extinction import calc_extinction_sphere
from .flip_ratio import \
    calc_iint, calc_flip_ratio_by_iint, \
    calc_asymmetry_by_iint

na = numpy.newaxis


def calc_symm_elem_points_by_index_points(index_point, n_abc):
    common_denominator = numpy.lcm.reduce(n_abc)
    coeff = numpy.floor_divide(numpy.lcm.reduce(n_abc), n_abc)
    denom = common_denominator*numpy.ones_like(index_point[0])
    symm_elem_points = numpy.stack([index_point[0]*coeff[0], index_point[1]*coeff[1], index_point[2]*coeff[2], denom], axis=0)
    return symm_elem_points


def calc_mem_col(
        index_hkl, unit_cell_parameters, eh_ccs, full_symm_elems, symm_elem_points,
        volume_unit_cell, number_unit_cell,
        point_multiplicity=None, dict_in_out: dict=None, flag_use_precalculated_data: bool = False):
    dict_in_out_keys = dict_in_out.keys()
    if (("index_hkl" in dict_in_out_keys) and flag_use_precalculated_data):
        if numpy.any(dict_in_out["index_hkl"] != index_hkl):
            flag_use_precalculated_data = False
            dict_in_out["index_hkl"] = index_hkl
    else:
        dict_in_out["index_hkl"] = index_hkl

    if (("eq_ccs" in dict_in_out_keys) and flag_use_precalculated_data):
        eq_ccs = dict_in_out["eq_ccs"]
    else:
        eq_ccs = calc_eq_ccs_by_unit_cell_parameters(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=False)[0]
        dict_in_out["eq_ccs"] = eq_ccs

    if point_multiplicity is None:
        if (("point_multiplicity" in dict_in_out_keys) and flag_use_precalculated_data):
            point_multiplicity = dict_in_out["point_multiplicity"]
        else:
            point_multiplicity = calc_multiplicity_by_atom_symm_elems(full_symm_elems, symm_elem_points)
            dict_in_out["point_multiplicity"] = point_multiplicity
    else:
        dict_in_out["point_multiplicity"] = point_multiplicity

    eh_perp_ccs = calc_vector_product_v1_v2_v1(eq_ccs, numpy.expand_dims(eh_ccs, axis=1), flag_v1=False, flag_v2=False)[0]
    fract_points = symm_elem_points[:3]/numpy.expand_dims(symm_elem_points[3], axis=0)

    rr = calc_m_v(
        numpy.expand_dims(full_symm_elems[4:13], axis=1),
        numpy.expand_dims(fract_points, axis=2), flag_m=False, flag_v=False)[0]

    fract_b = full_symm_elems[:3]/numpy.expand_dims(full_symm_elems[3], axis=0)
    hh = numpy.expand_dims(rr + numpy.expand_dims(fract_b, axis=1), axis=1)
    index_hkl_3d = numpy.expand_dims(numpy.expand_dims(index_hkl, axis=2), axis=3)
    # [hkl, points, symm]
    phase_3d = numpy.exp(-2.*numpy.pi * 1j*(hh[0] * index_hkl_3d[0] + hh[1] * index_hkl_3d[1] + hh[2] * index_hkl_3d[2]))
    phase_2d = phase_3d.sum(axis=2)
    # [3, hkl, points]
    mem_col = 0.2695*(numpy.expand_dims(phase_2d*numpy.expand_dims(point_multiplicity, axis=0),axis=0) * 
        numpy.expand_dims(eh_perp_ccs, axis=2))/full_symm_elems.shape[1]*volume_unit_cell/number_unit_cell
    return mem_col


def calc_mem_chi(
        index_hkl, unit_cell_parameters, h_ccs, full_symm_elems, symm_elem_points,
        point_susceptibility, volume_unit_cell, number_unit_cell,
        point_multiplicity=None, dict_in_out: dict=None, flag_use_precalculated_data: bool = False):
    dict_in_out_keys = dict_in_out.keys()
    if (("index_hkl" in dict_in_out_keys) and flag_use_precalculated_data):
        if numpy.any(dict_in_out["index_hkl"] != index_hkl):
            flag_use_precalculated_data = False
            dict_in_out["index_hkl"] = index_hkl
    else:
        dict_in_out["index_hkl"] = index_hkl

    if (("eq_ccs" in dict_in_out_keys) and flag_use_precalculated_data):
        eq_ccs = dict_in_out["eq_ccs"]
    else:
        eq_ccs = calc_eq_ccs_by_unit_cell_parameters(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=False)[0]
        dict_in_out["eq_ccs"] = eq_ccs

    if point_multiplicity is None:
        if (("point_multiplicity" in dict_in_out_keys) and flag_use_precalculated_data):
            point_multiplicity = dict_in_out["point_multiplicity"]
        else:
            point_multiplicity = calc_multiplicity_by_atom_symm_elems(full_symm_elems, symm_elem_points)
            dict_in_out["point_multiplicity"] = point_multiplicity
    else:
        dict_in_out["point_multiplicity"] = point_multiplicity

    m_r_inv_m = calc_m_r_inv_m(unit_cell_parameters, full_symm_elems, flag_unit_cell_parameters=False)[0]
    point_susceptibility_2d_ccs = calc_m1_m2_inv_m1(
        numpy.expand_dims(m_r_inv_m, axis=1),
        numpy.expand_dims(point_susceptibility, axis=2), flag_m1=False, flag_m2=False)[0]

    m_ccs = calc_m_v(point_susceptibility_2d_ccs, 
        numpy.expand_dims(numpy.expand_dims(h_ccs, axis=1), axis=2), flag_m=False, flag_v=False)[0]

    m_perp_ccs = calc_vector_product_v1_v2_v1(
        numpy.expand_dims(numpy.expand_dims(eq_ccs, axis=2), axis=3), 
        numpy.expand_dims(m_ccs, axis=1), flag_v1=False, flag_v2=False)[0]


    fract_points = symm_elem_points[:3]/numpy.expand_dims(symm_elem_points[3], axis=0)
    rr = calc_m_v(
        numpy.expand_dims(full_symm_elems[4:13], axis=1),
        numpy.expand_dims(fract_points, axis=2), flag_m=False, flag_v=False)[0]

    fract_b = full_symm_elems[:3]/numpy.expand_dims(full_symm_elems[3], axis=0)
    hh = numpy.expand_dims(rr + numpy.expand_dims(fract_b, axis=1), axis=1)
    index_hkl_3d = numpy.expand_dims(numpy.expand_dims(index_hkl, axis=2), axis=3)
    # [hkl, points, symm]
    phase_3d = numpy.exp(-2.*numpy.pi * 1j*(hh[0] * index_hkl_3d[0] + hh[1] * index_hkl_3d[1] + hh[2] * index_hkl_3d[2]))

    # [3, hkl, points]
    mem_chi = 0.2695*((numpy.expand_dims(phase_3d, axis=0) * m_perp_ccs).sum(axis=3))/full_symm_elems.shape[1] * \
        numpy.expand_dims(numpy.expand_dims(point_multiplicity, axis=0), axis=1)*volume_unit_cell/number_unit_cell
    return mem_chi


def renormailize_density_col(
        density_col, point_multiplicity,
        volume_unit_cell, points_unit_cell):
    coeff = float(points_unit_cell)/volume_unit_cell
    prod_den_mult = density_col*point_multiplicity
    norm_density_col = coeff*density_col/numpy.expand_dims(prod_den_mult.sum(axis=1), axis=1)
    return norm_density_col


def get_uniform_density_col(point_multiplicity, volume_unit_cell, points_unit_cell):
    density_col = numpy.ones((2, ) + point_multiplicity.shape, dtype=float)
    norm_density_col = renormailize_density_col(
        density_col, point_multiplicity, volume_unit_cell, points_unit_cell)
    return norm_density_col


def renormailize_density_chi(
        point_density, point_multiplicity, point_atom_label,
        point_atom_multiplicity,
        volume_unit_cell, points_unit_cell):
    na = numpy.newaxis
    coeff = float(points_unit_cell)/volume_unit_cell
    point_norm_density = numpy.copy(point_density)
    prod_den_mult = point_density*point_multiplicity/point_atom_multiplicity
    #FIXME: very slow solution. Redo it.
    atom_label = numpy.unique(point_atom_label)
    flag_2d = point_atom_label[na, :] == atom_label[:, na]

    for flag_1d in flag_2d:
        m_at = numpy.sum(prod_den_mult, where=flag_1d)
        if numpy.any(flag_1d):
            point_norm_density[flag_1d] *= coeff/float(m_at)
    return point_norm_density


def get_uniform_density_chi(point_multiplicity, point_atom_label, point_atom_multiplicity, volume_unit_cell, points_unit_cell):
    density_chi = numpy.ones(point_multiplicity.shape, dtype=float)
    norm_density_col = renormailize_density_chi(
        density_chi, point_multiplicity, point_atom_label,
        point_atom_multiplicity,
        volume_unit_cell, points_unit_cell)
    return norm_density_col


def save_spin_density_into_file(
        f_name: str, index_auc, spin_density, n_abc,
        unit_cell_parameters,
        reduced_symm_elems, translation_elems, centrosymmetry: bool, centrosymmetry_position):
    """Save to file.
    if centrosymmetry is False then centrosymmetry_position can be None
    """
    ls_out = []
    ls_out.append("Collinear density (created by CrysPy)")
    ls_out.append(f"{index_auc.shape[1]:}")

    for ind_xyz, sd in zip(index_auc.transpose(), spin_density.transpose()):
        ls_out.append(f"{ind_xyz[0]:4} {ind_xyz[1]:4} {ind_xyz[2]:4} {sd[0]+sd[1]:15.7f} {sd[0]:15.7f} {sd[1]:15.7f}")

    irad = 180./numpy.pi
    a, b = unit_cell_parameters[0], unit_cell_parameters[1]
    c, al = unit_cell_parameters[2], unit_cell_parameters[3]*irad
    be, ga = unit_cell_parameters[4]*irad, unit_cell_parameters[5]*irad

    ls_out.append(
        f"{a:10.5f}{b:10.5f}{c:10.5f}{al:10.5f}{be:10.5f}{ga:10.5f}")

    n_rs = reduced_symm_elems.shape[1]
    n_transl = translation_elems.shape[1]
    centr = int(centrosymmetry)
    ls_out.append(f"{n_abc[0]:5}{n_abc[1]:5}{n_abc[2]:5}{n_rs:5}{centr:5}{n_transl:5}")

    for s_e in reduced_symm_elems.transpose():
        r_11, r_12, r_13 = s_e[4], s_e[5], s_e[6]
        r_21, r_22, r_23 = s_e[7], s_e[8], s_e[9]
        r_31, r_32, r_33 = s_e[10], s_e[11], s_e[12]
        b_1, b_2, b_3 = s_e[0]/s_e[3], s_e[1]/s_e[3], s_e[2]/s_e[3]

        ls_out.append(
            f"{r_11:4}{r_21:4}{r_31:4}  {r_12:4}{r_22:4}{r_32:4}  \
{r_13:4}{r_23:4}{r_33:4}    {b_1:8.5f}{b_2:8.5f}{b_3:8.5f}")

    for orig in translation_elems.transpose():
        ls_out.append(
            f"{float(orig[0]/orig[3]):8.4f}{float(orig[1]/orig[3]):8.4f}\
{float(orig[2]/orig[3]):8.4f}")

    with open(f_name, "w") as fid:
        fid.write("\n".join(ls_out))

    return


def calc_index_atom_symmetry_closest_to_fract_xyz(
        fract_xyz, fract_atom_xyz, full_symm_elems, unit_cell_parameters):
    """
    Calculate index of atoms and applied symmetry to have closest atoms.

    Basins are defined as closest points to atom

    fract_xyz = [3, n_points]
    fract_atom_xyz = [3, n_atoms]

    Output:
        - n_atom_index = [n_points]: integers from 0 until n_atoms
        - n_symmetry = [n_points]: integers from 0 until n_symmetry
    """
    
    fract_xyz = numpy.mod(fract_xyz, 1.)
    fract_atom_xyz = fract_atom_xyz
    
    elem_r = full_symm_elems[4:13]
    fract_b = full_symm_elems[:3]/numpy.expand_dims(full_symm_elems[3], axis=0)
    rr_atom = calc_m_v(
        numpy.expand_dims(elem_r, axis=1),
        numpy.expand_dims(fract_atom_xyz, axis=2), flag_m=False, flag_v=False)[0]
    
    fract_atom_s = numpy.mod(
        rr_atom + numpy.expand_dims(fract_b, axis=1), 1.)
    
    diff_fract = (numpy.expand_dims(numpy.expand_dims(fract_xyz, axis=2), axis=3) -
                  numpy.expand_dims(fract_atom_s, axis=1))

    flag_g_half = numpy.abs(diff_fract) > 0.5
    flag_sign = numpy.where(diff_fract >= 0.0, -1.0, 1.0)
    diff_fract_1 = numpy.where(flag_g_half, diff_fract+flag_sign, diff_fract)
    
    m_m = calc_m_m_by_unit_cell_parameters(unit_cell_parameters, flag_unit_cell_parameters=False)[0]
    diff_position_1 = calc_m_v(m_m, diff_fract_1, flag_m=False, flag_v=False)[0]

    dist_sq_3d = numpy.square(diff_position_1).sum(axis=0)
    dist_sq_2d_over_sym = numpy.min(dist_sq_3d, axis=2)
    dist_sq_2d_over_at = numpy.min(dist_sq_3d, axis=1)
    
    ind_at = numpy.argmin(dist_sq_2d_over_sym, axis=1)
    ind_sym = numpy.argmin(dist_sq_2d_over_at, axis=1)
    dist_sq = numpy.min(dist_sq_2d_over_sym, axis=1)
    distance = numpy.sqrt(dist_sq)

    return ind_at, ind_sym, distance

def form_basins(
        symm_elem_auc, 
        full_symm_elems, unit_cell_parameters,
        atom_label, atom_fract_xyz, atom_multiplicity, mag_atom_label):

    # fractions in asymmetric unit cell
    point_fract_xyz_auc = symm_elem_auc[:3, :]/numpy.expand_dims(symm_elem_auc[3, :], axis=0)

    # separate on basins
    ind_at, ind_sym, point_atom_distance = calc_index_atom_symmetry_closest_to_fract_xyz(
            point_fract_xyz_auc, atom_fract_xyz, full_symm_elems, unit_cell_parameters)


    point_atom_label = atom_label[ind_at]
    point_atom_multiplicity = atom_multiplicity[ind_at]

    point_atom_symm_elems = full_symm_elems[:, ind_sym]

    flag_chi = numpy.sum(
        numpy.expand_dims(point_atom_label, axis=1) == numpy.expand_dims(mag_atom_label, axis=0),
        axis=1).astype(bool)

    atom_label_auc_chi = point_atom_label[flag_chi]
    atom_multiplicity_auc_chi = point_atom_multiplicity[flag_chi]
    atom_distance_auc_chi = point_atom_distance[flag_chi]
    atom_symm_elems_auc_chi = point_atom_symm_elems[:, flag_chi]

    return flag_chi, atom_label_auc_chi, atom_multiplicity_auc_chi, atom_distance_auc_chi, atom_symm_elems_auc_chi

def calc_point_susceptibility(
        unit_cell_parameters, atom_symm_elems_chi, atom_label_chi, atom_para_label, atom_para_susceptibility, atom_para_sc_chi,
        full_symm_elems, point_symm_elem):
    atom_para_susceptibility_averaged = (atom_para_sc_chi * numpy.expand_dims(atom_para_susceptibility,axis=0)).sum(axis=1)
    m_r_inv_m = calc_m_r_inv_m(unit_cell_parameters, atom_symm_elems_chi, flag_unit_cell_parameters=False)[0]
    flag_mag = numpy.expand_dims(atom_label_chi, axis=1) == numpy.expand_dims(atom_para_label, axis=0)
    ind_mag_atom = numpy.argmax(flag_mag, axis=1)
    point_susceptibility_a = atom_para_susceptibility_averaged[:, ind_mag_atom]
    point_susceptibility = calc_m_q_inv_m(m_r_inv_m, point_susceptibility_a, flag_m=False, flag_q=False)[0]

    flag_2d = calc_symm_flags(
        numpy.expand_dims(full_symm_elems, axis=2), 
        numpy.expand_dims(point_symm_elem, axis=1))

    l_scm_chi = []
    for ind in range(flag_2d.shape[1]):
        symm_elem_direct = full_symm_elems[:, flag_2d[:, ind]]
        # r_ccs = calc_m_r_inv_m(unit_cell_parameters, symm_elem_direct, flag_unit_cell_parameters=False)[0]
        scm_chi = calc_sc_chi_full(symm_elem_direct, unit_cell_parameters, flag_unit_cell_parameters=False)[0]
        l_scm_chi.append(scm_chi)
    point_scm_chi = numpy.stack(l_scm_chi, axis=-1)

    point_susceptibility_aver = (point_scm_chi * numpy.expand_dims(point_susceptibility,axis=0)).sum(axis=1)
    return point_susceptibility_aver


def calc_chi_atoms(
        unit_cell_parameters, number_points, full_symm_elems,
        index_hkl, atom_para_fract_xyz, atom_para_sc_chi, symm_elem_channel_chi,
        point_multiplicity_channel_chi, density_channel_chi):
    
    m_r_inv_m, dder = calc_m_r_inv_m(unit_cell_parameters, full_symm_elems)
    mim_1, dder = calc_mm_as_m1_m2_inv_m1(m_r_inv_m)
    m_m, dder = calc_m_m_by_unit_cell_parameters(unit_cell_parameters)
    
    
    # for SCM
    flag_2d = calc_symm_flags(
        numpy.expand_dims(full_symm_elems, axis=2), 
        numpy.expand_dims(symm_elem_channel_chi, axis=1))
    
    l_scm_chi = []
    for ind in range(flag_2d.shape[1]):
        symm_elem_direct = full_symm_elems[:, flag_2d[:, ind]]
        scm_chi = calc_sc_chi_full(symm_elem_direct, unit_cell_parameters, flag_unit_cell_parameters=False)[0]
        l_scm_chi.append(scm_chi)
    point_scm_chi = numpy.stack(l_scm_chi, axis=-1)
    
    point_fract_xyz_auc = symm_elem_channel_chi[:3, :]/numpy.expand_dims(symm_elem_channel_chi[3, :], axis=0)
    ind_at, ind_sym, point_atom_distance = calc_index_atom_symmetry_closest_to_fract_xyz(
            point_fract_xyz_auc, atom_para_fract_xyz, full_symm_elems, unit_cell_parameters)
    
    symm_elems_ai = full_symm_elems[:, ind_sym]
    
    m_r_ai_inv_m, dder = calc_m_r_inv_m(unit_cell_parameters, symm_elems_ai)
    mim_2, dder = calc_mm_as_m_q_inv_m(m_r_ai_inv_m)
    
    rr = calc_m_v(
        numpy.expand_dims(full_symm_elems[4:13], axis=1),
        numpy.expand_dims(point_fract_xyz_auc, axis=2), flag_m=False, flag_v=False)[0]

    fract_b = full_symm_elems[:3]/numpy.expand_dims(full_symm_elems[3], axis=0)
    hh = numpy.expand_dims(rr + numpy.expand_dims(fract_b, axis=1), axis=1)
    index_hkl_3d = numpy.expand_dims(numpy.expand_dims(index_hkl, axis=2), axis=3)
    # [hkl, points, symm]
    phase_3d = numpy.exp(-2.*numpy.pi * 1j*(hh[0] * index_hkl_3d[0] + hh[1] * index_hkl_3d[1] + hh[2] * index_hkl_3d[2]))    
    
    # sum over symmetry
    l_hhh = [(numpy.expand_dims(phase_3d[ind:(ind+1), :, :], axis=(0,1)) *
          numpy.expand_dims(mim_1, axis=(2,3))).sum(axis=4) for ind in range(phase_3d.shape[0])]
    
    hhh = numpy.concatenate(l_hhh, axis=2)
    
    hhh2 = numpy.expand_dims(point_multiplicity_channel_chi*density_channel_chi, axis=(0,1,2)) * hhh 
    
    
    hh1 = (numpy.expand_dims(mim_2, axis=(2,3)) * numpy.expand_dims(atom_para_sc_chi, axis=(0,4))).sum(axis=1)
    hh2 = (numpy.expand_dims(point_scm_chi, axis=(2,3)) * numpy.expand_dims(hh1, axis=0)).sum(axis=1)
    
    # sum over nu
    l_flag_at = [ind_at == i_at for i_at in range(ind_at.max()+1)]
    l_hhh3 = []
    for ind in range(hhh2.shape[2]):
        hhhh3 = ((numpy.expand_dims(hhh2[:, :, ind:(ind+1), :], axis=(3, 4)) * 
                  numpy.expand_dims(hh2, axis=(0, 2)))).sum(axis=1)
        l_hhh4 = [(hhhh3[:, :, :, ind_at, :]*numpy.expand_dims(flag_at, axis=(0,1,2))).sum(axis=3) for ind_at, flag_at in enumerate(l_flag_at)]
        hhh3 = numpy.stack(l_hhh4, axis=3)
        l_hhh3.append(hhh3)
    hhh3 = numpy.concatenate(l_hhh3, axis=1)    
    
    
    volume_uc, dder = calc_volume_uc_by_unit_cell_parameters(unit_cell_parameters)
    coeff = 0.2695 * volume_uc/(full_symm_elems.shape[1]*number_points)
    chi_atoms = coeff*hhh3
    return chi_atoms


def calc_f_m_perp_ccs_by_chi_atoms(chi_atoms, magnetic_field, vp, atom_para_susceptibility):
    m_chi = (chi_atoms*numpy.expand_dims(atom_para_susceptibility, axis=(0,1))).sum(axis=(2,3))
    mv, dder = calc_m_v(m_chi, magnetic_field)
    f_m_perp_chi = (vp * numpy.expand_dims(mv, axis=0)).sum(axis=1)
    return f_m_perp_chi

def calc_model_value_by_precalculated_data(atom_para_susceptibility, unit_cell_parameters, flag_asymmetry, dict_in_out, l_dict_diffrn):
    flag_use_precalculated_data = False
    volume_unit_cell, dder = calc_volume_uc_by_unit_cell_parameters(unit_cell_parameters)
    l_model_value = []
    for dict_diffrn in l_dict_diffrn:
        diffrn_dict_in_out = dict_in_out["dict_in_out_"+dict_diffrn['type_name']]
        vp = diffrn_dict_in_out["vp"]
        chi_atoms = diffrn_dict_in_out["chi_atoms"]
    
        f_nucl = diffrn_dict_in_out["f_nucl"]
    
        index_hkl = dict_diffrn["index_hkl"]
        magnetic_field = dict_diffrn["magnetic_field"]

        beam_polarization = dict_diffrn["beam_polarization"]
        flipper_efficiency = dict_diffrn["flipper_efficiency"]
        matrix_u = dict_diffrn["matrix_u"]
        flip_ratio_es = dict_diffrn["flip_ratio_es"]
    
        f_m_perp_plus_minus = numpy.zeros(index_hkl.shape, dtype=complex)

    
        f_m_perp_chi = calc_f_m_perp_ccs_by_chi_atoms(chi_atoms, magnetic_field, vp, atom_para_susceptibility)
    
        f_m_perp = f_m_perp_chi + f_m_perp_plus_minus


        wavelength = dict_diffrn["wavelength"]
        sthovl = calc_sthovl_by_unit_cell_parameters(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=False)[0]
        cos_2theta = numpy.cos(2*numpy.arcsin(sthovl*wavelength))

        extinction_model = dict_diffrn["extinction_model"]
        extinction_radius = dict_diffrn["extinction_radius"]
        extinction_mosaicity = dict_diffrn["extinction_mosaicity"]        
        
        func_extinction = lambda f_sq, flag_f_sq: calc_extinction_sphere(
            f_sq, extinction_radius, extinction_mosaicity, volume_unit_cell, cos_2theta, wavelength,
            extinction_model, flag_f_sq=False, flag_radius=False,
            flag_mosaicity=False,
            flag_volume_unit_cell=False,
            flag_cos_2theta=False,
            flag_wavelength=False)

        axis_z = matrix_u[6:9]
        iint_plus, iint_minus, dder_plus, dder_minus = calc_iint(
            beam_polarization, flipper_efficiency, f_nucl, f_m_perp, axis_z, func_extinction = func_extinction,
            flag_beam_polarization = False, flag_flipper_efficiency = False,
            flag_f_nucl = False, flag_f_m_perp = True,
            dict_in_out = dict_in_out, flag_use_precalculated_data = flag_use_precalculated_data)

    
        if flag_asymmetry:
            model_exp, dder_model_exp = calc_asymmetry_by_iint(
                iint_plus, iint_minus, c_lambda2=None, iint_2hkl=None,
                flag_iint_plus=True, flag_iint_minus=True, 
                flag_c_lambda2=False, flag_iint_2hkl=False)
        else:
            model_exp, dder_model_exp = calc_flip_ratio_by_iint(
                iint_plus, iint_minus, c_lambda2=None, iint_2hkl=None,
                flag_iint_plus=True, flag_iint_minus=True, 
                flag_c_lambda2=False, flag_iint_2hkl=False)
        l_model_value.append(model_exp)
    model_value = numpy.concatenate(l_model_value, axis=0)
    return model_value





#FIXME delete
def calc_point_chi_direct(
        mag_atom_label, atom_chi, symm_elems, symm_flag_point_to_point,
        point_atom_label, point_atom_symm_elems, unit_cell_parameters):
    """Calculate susceptibility of point i in direct unit cell.
    mag_atom_label[n_mag_atom] is label of magnetic atoms
    atom_chi[6, n_mag_atom] is local susceptibility of magnetic atoms
    symm_elems[13, n_el_symm] is symmetry operations
    symm_flag_point_to_point[] is flag of symmetry operations which transform point to point
    point_atom_label[n_point] is the label of atom to whom the point is belongs
    point_atom_symm_elems[n_point, ] is the symmetry elements 

    Output:
    point_chi_direct[6, n_point] is the susceptibility parameters in direct unit cell
    """

    flag_2d = point_atom_label[:, na] == mag_atom_label[na, :]
    index_susc = numpy.argmax(flag_2d, axis=1)
    chi_recip_norm = atom_chi[:, index_susc]
    p_a_s_el_r = point_atom_symm_elems[4:, :]

    chi_s_direct, dder_ = calc_chi_direct_norm_with_symmetry(chi_recip_norm, p_a_s_el_r)

    symm_elems_r = symm_elems[4:, :]
    chi_s_direct_2d = calc_m1_m2_inv_m1(symm_elems_r[:, :, na], chi_s_direct[:, na, :], flag_m1=False, flag_m2=False)[0]
    hh = symm_flag_point_to_point.astype(int)
    n_at = (hh.sum(axis=0)).astype(float)
    point_chi_direct = (chi_s_direct_2d*hh[na, :, :]).sum(axis=1)/n_at
    
    return point_chi_direct


#FIXME delete
def calc_point_chi_ccs(
        mag_atom_label, atom_chi, symm_elems, symm_flag_point_to_point,
        point_atom_label, point_atom_symm_elems, unit_cell_parameters):
    """Calculate susceptibility of point i in cartesian coordinate system.
    mag_atom_label[n_mag_atom] is label of magnetic atoms
    atom_chi[6, n_mag_atom] is local susceptibility of magnetic atoms
    symm_elems[13, n_el_symm] is symmetry operations
    symm_flag_point_to_point[] is flag of symmetry operations which transform point to point
    point_atom_label[n_point] is the label of atom to whom the point is belongs
    point_atom_symm_elems[n_point, ] is the symmetry elements 

    Output:
    point_chi_direct[6, n_point] is the susceptibility parameters in cartesian coordinate system
    """

    point_chi_ccs = calc_point_chi_direct(
        mag_atom_label, atom_chi, symm_elems, symm_flag_point_to_point,
        point_atom_label, point_atom_symm_elems, unit_cell_parameters)
    # point_chi_ccs, dder_ = calc_susceptibility_ccs(point_chi_direct, unit_cell_parameters)
    return point_chi_ccs


#FIXME delete
def calc_point_moment_ccs_by_point_chi_direct(symm_elems, point_chi_ccs, field_ccs, unit_cell_parameters):
    """Calculate magnetic moments [3, n_points, n_symm_elems]
    """
    chi_s_ccs_2d = calc_m1_m2_inv_m1(symm_elems[4:, :, na], point_chi_ccs[:, na, :], flag_m1=False, flag_m2=False)[0]
    # chi_s_ccs_2d, dder_ = calc_susceptibility_ccs(chi_s_direct_2d, unit_cell_parameters)
    moment_ccs = calc_m_v(chi_s_ccs_2d, field_ccs, flag_m=False, flag_v=False)[0]
    return moment_ccs





import numpy
import scipy
import scipy.interpolate
import scipy.special

from cryspy.A_functions_base.matrix_operations import calc_m1_m2_m1t

from cryspy.A_functions_base.unit_cell import \
    calc_sthovl_by_unit_cell_parameters, calc_matrix_t

from cryspy.A_functions_base.structure_factor import \
    calc_f_nucl_by_dictionary, calc_sft_ccs_by_dictionary, \
    calc_index_hkl_multiplicity_in_range

from cryspy.A_functions_base.integrated_intensity_powder_diffraction import \
    calc_powder_iint_1d_para

from cryspy.A_functions_base.preferred_orientation import calc_preferred_orientation_pd

from cryspy.A_functions_base.powder_diffraction_const_wavelength import \
    calc_lorentz_factor

from cryspy.A_functions_base.powder_diffraction_tof import \
    calc_spectrum, calc_time_for_epithermal_neutrons_by_d, calc_time_for_thermal_neutrons_by_d, \
    calc_d_by_time_for_thermal_neutrons, calc_d_min_max_by_time_thermal_neutrons, \
    calc_d_min_max_by_time_epithermal_neutrons, calc_peak_shape_function

from cryspy.A_functions_base.powder_diffraction_tof_zcode import \
    calc_profile_by_zcode_parameters

from .rhochi_diffrn import get_flags
from .rhochi_pd import calc_background

na = numpy.newaxis


def calc_background_by_cosines(x, background_coefficients, x_min: float = 0., x_max: float = None, flag_background_coefficients: bool = False):
    if x_max is None:
        x_max = numpy.max(x)
    x_rel = ((x-x_min)/x_max)*numpy.pi
    val = numpy.linspace(
        0, background_coefficients.shape[0]-1, background_coefficients.shape[0])
    res = numpy.cos(numpy.expand_dims(val, axis=1) *
                    numpy.expand_dims(x_rel, axis=0))
    y = numpy.sum(numpy.expand_dims(
        background_coefficients, axis=1) * res, axis=0)
    dder = {}
    if flag_background_coefficients:
        dder["background_coefficients"] = res
    return y, dder


def calc_spectrum_incident(time, coefficients, type: str = "Maxwell", flag_coefficients: bool = False):
    exp = numpy.exp
    time_sq = numpy.square(time)
    time_4 = numpy.square(time_sq)
    a0, a1, a2, a3, a4 = coefficients[0], coefficients[1], coefficients[2], coefficients[3], coefficients[4]
    a5, a6, a7, a8 = coefficients[5], coefficients[6], coefficients[7], coefficients[8]
    if type == "Empirical-Exponents":
        res = a0 + a1 * exp(-a2 * time) + \
            a3 * exp(-a4 * time_sq) + \
            a5 * exp(-a6 * time*time_sq) + \
            a7 * exp(-a8 * time_4)
    else:  # type == "Maxwell"
        res = a0 + a1 * exp(-a2 * time_sq)/(time*time_4) + \
            a3 * exp(-a4 * time_sq) + \
            a5 * exp(-a6 * time*time_sq) + \
            a7 * exp(-a8 * numpy.square(time_sq))

    dder = {}
    if flag_coefficients:
        dder["coefficients"] = None
    return res, dder


def calc_chi_sq_for_tof_by_dictionary(
        dict_tof, dict_crystals, dict_in_out: dict = None, flag_use_precalculated_data: bool = False,
        flag_calc_analytical_derivatives: bool = False):
    """Calculate chi_sq for diffrn experiment.
    """
    if dict_in_out is None:
        flag_dict = False
        flag_use_precalculated_data = False
        dict_in_out_keys = []
    else:
        flag_dict = True
        dict_in_out_keys = dict_in_out.keys()
    dict_tof_keys = dict_tof.keys()

    phase_name = [hh["name"].lower() for hh in dict_crystals]

    excluded_points = dict_tof["excluded_points"]
    time = dict_tof["time"]
    dict_in_out["time"] = time
    neutron_type = dict_tof["neutron_type"]
    if neutron_type == "thermal":
        zero = dict_tof["zero"]
        dtt1 = dict_tof["dtt1"]
        dtt2 = dict_tof["dtt2"]
        d = calc_d_by_time_for_thermal_neutrons(time, zero, dtt1, dtt2)
        d_min_max = calc_d_min_max_by_time_thermal_neutrons(
            time, zero, dtt1, dtt2)
        flag_time = (
            numpy.any(dict_tof["flags_zero"]) or
            numpy.any(dict_tof["flags_dtt1"]) or
            numpy.any(dict_tof["flags_dtt2"]))
    else:  # epithermal
        zero = dict_tof["zero"]
        dtt1 = dict_tof["dtt1"]
        zerot = dict_tof["zerot"]
        dtt1t = dict_tof["dtt1t"]
        dtt2t = dict_tof["dtt2t"]
        flag_time = (
            numpy.any(dict_tof["flags_zero"]) or
            numpy.any(dict_tof["flags_dtt1"]) or
            numpy.any(dict_tof["flags_zerot"]) or
            numpy.any(dict_tof["flags_dtt1t"]) or
            numpy.any(dict_tof["flags_dtt2t"]))
        d_min_max = calc_d_min_max_by_time_epithermal_neutrons(
            time, zero, dtt1, zerot, dtt1t, dtt2t)
        raise AttributeError("Epithermal neutrons are not introudiced")

    sthovl_min = 0.5/d_min_max[1]
    sthovl_max = 0.5/d_min_max[0]

    if flag_dict:
        dict_in_out["d"] = d
        dict_in_out["excluded_points"] = excluded_points

    ttheta_bank = dict_tof["ttheta_bank"]
    lorentz_factor = calc_lorentz_factor(ttheta_bank, flag_ttheta=False)[0]

    if "beam_polarization" in dict_tof_keys:
        beam_polarization = dict_tof["beam_polarization"]
        flipper_efficiency = dict_tof["flipper_efficiency"]
        magnetic_field = dict_tof["magnetic_field"]
        flags_beam_polarization = dict_tof["flags_beam_polarization"]
        flags_flipper_efficiency = dict_tof["flags_flipper_efficiency"]
    else:
        beam_polarization, flipper_efficiency, magnetic_field = 0., 0., 0.
        flags_beam_polarization, flags_flipper_efficiency = False, False

    dict_tof_keys = dict_tof.keys()
    if "background_coefficients" in dict_tof_keys:
        background_coefficients = dict_tof["background_coefficients"]
        flags_background_coefficients = dict_tof["flags_background_coefficients"]

        flag_background_coefficients = numpy.any(flags_background_coefficients)
        if (flag_use_precalculated_data and ("signal_background" in dict_in_out) and not(flag_background_coefficients)):
            signal_background = dict_in_out["signal_background"]
        else:
            signal_background, dder_s_bkgr = calc_background_by_cosines(time, background_coefficients,
                                                                        flag_background_coefficients=(flag_background_coefficients and flag_calc_analytical_derivatives))
            dict_in_out["signal_background"] = signal_background
    elif "background_time" in dict_tof_keys:
        background_time = dict_tof["background_time"]
        background_intensity = dict_tof["background_intensity"]
        flags_background_intensity = dict_tof["flags_background_intensity"]
        flag_background_intensity = numpy.any(flags_background_intensity)
        if (flag_use_precalculated_data and ("signal_background" in dict_in_out) and not(flag_background_intensity)):
            signal_background = dict_in_out["signal_background"]
        else:
            signal_background, dder_s_bkgr = calc_background(
                time,
                background_time,
                background_intensity,
                flag_background_intensity=(flag_background_intensity and flag_calc_analytical_derivatives))
            dict_in_out["signal_background"] = signal_background
    else:
        signal_background = numpy.zeros_like(time)
        dict_in_out["signal_background"] = signal_background

    if "spectrum_incident_type" in dict_tof_keys:
        spectrum_incident_type = dict_tof["spectrum_incident_type"]
        spectrum_incident_coefficients = dict_tof["spectrum_incident_coefficients"]
        flags_spectrum_incident_coefficients = dict_tof["flags_spectrum_incident_coefficients"]
        flag_spectrum_incident_coefficients = numpy.any(
            flags_spectrum_incident_coefficients)
        if (flag_use_precalculated_data and ("spectrum_incident" in dict_in_out) and not(flag_spectrum_incident_coefficients)):
            spectrum_incident = dict_in_out["spectrum_incident"]
        else:
            spectrum_incident, dder_si = calc_spectrum_incident(
                time, spectrum_incident_coefficients, type=spectrum_incident_type)
        dict_in_out["spectrum_incident"] = spectrum_incident
    else:
        spectrum_incident = numpy.ones_like(time)
        dict_in_out["spectrum_incident"] = spectrum_incident

    phase_name = [hh["name"].lower() for hh in dict_crystals]

    pd_phase_name = dict_tof["phase_name"]
    phase_scale = dict_tof["phase_scale"]

    phase_ig = dict_tof["phase_ig"]  # IG_phase
    flags_phase_scale = dict_tof["flags_phase_scale"]
    flags_phase_ig = dict_tof["flags_phase_ig"]  # IG_phase

    profile_peak_shape = dict_tof["profile_peak_shape"]
    if profile_peak_shape == "pseudo-Voigt":
        profile_alphas = dict_tof["profile_alphas"]
        profile_betas = dict_tof["profile_betas"]
        profile_sigmas = dict_tof["profile_sigmas"]
        profile_gammas = dict_tof["profile_gammas"]
        flag_profile_shape = (
            numpy.any(dict_tof["flags_profile_alphas"]) or
            numpy.any(dict_tof["flags_profile_betas"]) or
            numpy.any(dict_tof["flags_profile_sigmas"]) or
            numpy.any(dict_tof["flags_profile_gammas"]))
    elif profile_peak_shape == "Gauss":
        profile_alphas = dict_tof["profile_alphas"]
        profile_betas = dict_tof["profile_betas"]
        profile_sigmas = dict_tof["profile_sigmas"]
        flag_profile_shape = (
            numpy.any(dict_tof["flags_profile_alphas"]) or
            numpy.any(dict_tof["flags_profile_betas"]) or
            numpy.any(dict_tof["flags_profile_sigmas"]))
    elif profile_peak_shape == "type0m":
        profile_alphas = dict_tof["profile_alphas"]
        profile_betas = dict_tof["profile_betas"]
        profile_sigmas = dict_tof["profile_sigmas"]
        profile_gammas = dict_tof["profile_gammas"]
        profile_rs = dict_tof["profile_rs"]
        flag_profile_shape = (
            numpy.any(dict_tof["flags_profile_alphas"]) or
            numpy.any(dict_tof["flags_profile_betas"]) or
            numpy.any(dict_tof["flags_profile_sigmas"]) or
            numpy.any(dict_tof["flags_profile_gammas"]) or
            numpy.any(dict_tof["flags_profile_rs"]))

    if "texture_name" in dict_tof_keys:
        flag_texture = True
        texture_name = dict_tof["texture_name"]
        texture_g1 = dict_tof["texture_g1"]
        texture_g2 = dict_tof["texture_g2"]
        texture_axis = dict_tof["texture_axis"]
        flags_texture_g1 = dict_tof["flags_texture_g1"]
        flags_texture_g2 = dict_tof["flags_texture_g2"]
        flags_texture_axis = dict_tof["flags_texture_axis"]
    else:
        flag_texture = False

    total_signal_plus = numpy.zeros_like(time)
    total_signal_minus = numpy.zeros_like(time)
    for p_name, p_scale, p_ig, flags_p_scale, flags_p_ig in zip(
            pd_phase_name, phase_scale, phase_ig, flags_phase_scale, flags_phase_ig):
        p_name = p_name.lower()
        flag_phase_texture = False
        if flag_texture:
            ind_texture = numpy.argwhere(texture_name == p_name)
            if ind_texture.shape[0] != 0:
                texture_g1 = texture_g1[ind_texture[0]]
                texture_g2 = texture_g2[ind_texture[0]]
                texture_axis = texture_axis[:, ind_texture[0]]
                flag_phase_texture = True
                flags_texture_g1 = flags_texture_g1[ind_texture[0]]
                flags_texture_g2 = flags_texture_g2[ind_texture[0]]
                flags_texture_axis = flags_texture_axis[:, ind_texture[0]]

        ind_phase = phase_name.index(p_name)
        dict_crystal = dict_crystals[ind_phase]
        dict_in_out_keys = dict_in_out.keys()
        if f"dict_in_out_{p_name:}" in dict_in_out_keys:
            dict_in_out_phase = dict_in_out[f"dict_in_out_{p_name:}"]
        else:
            dict_in_out_phase = {}
            dict_in_out[f"dict_in_out_{p_name:}"] = dict_in_out_phase

        dict_in_out_phase_keys = dict_in_out_phase.keys()

        reduced_symm_elems = dict_crystal["reduced_symm_elems"]
        translation_elems = dict_crystal["translation_elems"]
        centrosymmetry = dict_crystal["centrosymmetry"]
        unit_cell_parameters = dict_crystal["unit_cell_parameters"]
        flags_unit_cell_parameters = dict_crystal["flags_unit_cell_parameters"]
        flag_unit_cell_parameters = numpy.any(flags_unit_cell_parameters)

        if flag_unit_cell_parameters:
            sc_uc = dict_crystal["sc_uc"]
            v_uc = dict_crystal["v_uc"]
            unit_cell_parameters = numpy.dot(
                sc_uc, unit_cell_parameters) + v_uc

        if (flag_use_precalculated_data and
                ("index_hkl" in dict_in_out_phase_keys) and
                ("multiplicity_hkl" in dict_in_out_phase_keys) and not(flag_unit_cell_parameters)):
            index_hkl = dict_in_out_phase["index_hkl"]
            multiplicity_hkl = dict_in_out_phase["multiplicity_hkl"]
        else:
            if flag_phase_texture:
                reduced_symm_elems_p1 = numpy.array([[0], [0], [0], [1], [1], [0], [0], [
                                                    0], [1], [0], [0], [0], [1]], dtype=int)
                translation_elems_p1 = numpy.array(
                    [[0], [0], [0], [1]], dtype=int)
                index_hkl, multiplicity_hkl = calc_index_hkl_multiplicity_in_range(
                    sthovl_min, sthovl_max, unit_cell_parameters, reduced_symm_elems_p1, translation_elems_p1, centrosymmetry)
            else:
                index_hkl, multiplicity_hkl = calc_index_hkl_multiplicity_in_range(
                    sthovl_min, sthovl_max, unit_cell_parameters, reduced_symm_elems, translation_elems, centrosymmetry)

            if (("index_hkl" in dict_in_out_phase_keys) and flag_use_precalculated_data):
                if index_hkl.shape != dict_in_out_phase["index_hkl"].shape:
                    flag_use_precalculated_data = False
                else:
                    flag_use_precalculated_data = numpy.all(
                        numpy.logical_and(dict_in_out_phase["index_hkl"], index_hkl))

            dict_in_out_phase["index_hkl"] = index_hkl
            dict_in_out_phase["multiplicity_hkl"] = multiplicity_hkl

        flag_sthovl_hkl = flag_unit_cell_parameters
        sthovl_hkl, dder_sthovl_hkl = calc_sthovl_by_unit_cell_parameters(index_hkl,
                                                                          unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)

        d_hkl = 0.5/sthovl_hkl
        if neutron_type == "thermal":
            time_hkl = calc_time_for_thermal_neutrons_by_d(
                d_hkl, zero, dtt1, dtt2)
        else:  # neutron_type == "epithermal"
            time_hkl = calc_time_for_epithermal_neutrons_by_d(
                d_hkl, zero, dtt1, zerot, dtt1t, dtt2t)

        wavelength_hkl = 2.*d_hkl*numpy.sin(0.5*ttheta_bank)
        wavelength_4_hkl = numpy.power(wavelength_hkl, 4)

        flag_time_hkl = flag_sthovl_hkl
        dict_in_out_phase["time_hkl"] = time_hkl
        dict_in_out_phase["d_hkl"] = d_hkl

        f_nucl, dder_f_nucl = calc_f_nucl_by_dictionary(
            dict_crystal, dict_in_out_phase, flag_use_precalculated_data=flag_use_precalculated_data)
        flag_f_nucl = len(dder_f_nucl.keys()) > 0

        sft_ccs, dder_sft_ccs = calc_sft_ccs_by_dictionary(
            dict_crystal, dict_in_out_phase, flag_use_precalculated_data=flag_use_precalculated_data)
        flag_sft_ccs = len(dder_sft_ccs.keys()) > 0

        flag_matrix_t = flag_unit_cell_parameters
        matrix_t, dder_matrix_t = calc_matrix_t(
            index_hkl, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)

        flag_tensor_sigma = flag_sft_ccs or flag_unit_cell_parameters
        tensor_sigma, dder_tensor_sigma = calc_m1_m2_m1t(
            matrix_t, sft_ccs, flag_m1=flag_sft_ccs, flag_m2=flag_unit_cell_parameters)

        flag_iint_plus_minus = flag_f_nucl or flag_tensor_sigma or flags_beam_polarization or flags_flipper_efficiency

        if (("iint_plus" in dict_in_out_phase_keys) and ("iint_minu" in dict_in_out_phase_keys) and
                flag_use_precalculated_data and not(flag_iint_plus_minus)):
            iint_plus, iint_minus = dict_in_out_phase["iint_plus"], dict_in_out_phase["iint_minus"]
        else:
            iint_plus, iint_minus, dder_plus, dder_minus = calc_powder_iint_1d_para(
                f_nucl, tensor_sigma, beam_polarization, flipper_efficiency, magnetic_field,
                flag_f_nucl=flag_f_nucl, flag_tensor_sigma=flag_tensor_sigma,
                flag_polarization=flags_beam_polarization, flag_flipper=flags_flipper_efficiency)
            dict_in_out_phase["iint_plus"] = iint_plus
            dict_in_out_phase["iint_minus"] = iint_minus

        if flag_phase_texture:
            flag_texture_g1 = numpy.any(flags_texture_g1)
            flag_texture_g2 = numpy.any(flags_texture_g2)
            flag_texture_axis = numpy.any(flags_texture_axis)
            flag_hh = numpy.any(
                [flag_texture_g1, flag_texture_g2, flag_texture_axis])
            if (flag_use_precalculated_data and
                    ("preferred_orientation" in dict_in_out_phase_keys) and
                    not(flag_hh)):
                preferred_orientation = dict_in_out_phase["preferred_orientation"]
            else:
                preferred_orientation, dder_po = calc_preferred_orientation_tof(
                    index_hkl, texture_g1, texture_g2, texture_axis, unit_cell_parameters,
                    flag_texture_g1=flag_texture_g1 and flag_calc_analytical_derivatives,
                    flag_texture_g2=flag_texture_g2 and flag_calc_analytical_derivatives,
                    flag_texture_axis=flag_texture_axis and flag_calc_analytical_derivatives)
                dict_in_out_phase["preferred_orientation"] = preferred_orientation

        flag_profile_tof = (
            flag_unit_cell_parameters or flag_time or flag_profile_shape)
        if (("profile_tof" in dict_in_out_phase_keys) and
                flag_use_precalculated_data and not(flag_profile_tof)):
            profile_tof = dict_in_out_phase["profile_tof"]
        else:
            if profile_peak_shape == "pseudo-Voigt":
                profile_tof = calc_peak_shape_function(
                    profile_alphas, profile_betas, profile_sigmas,
                    d, time, time_hkl,
                    gammas=profile_gammas, size_g=0., size_l=0.,
                    strain_g=0., strain_l=0., peak_shape=profile_peak_shape)
            elif profile_peak_shape == "Gauss":
                profile_tof = calc_peak_shape_function(
                    profile_alphas, profile_betas, profile_sigmas,
                    d, time, time_hkl,
                    gammas=None, size_g=0., size_l=0.,
                    strain_g=0., strain_l=0., peak_shape=profile_peak_shape)
            elif profile_peak_shape == "type0m":
                time_2d = numpy.expand_dims(time, axis=1)
                time_hkl_2d = numpy.expand_dims(time_hkl, axis=0)
                d_hkl_2d = numpy.expand_dims(d_hkl, axis=0)
                delta_t_2d = time_2d - time_hkl_2d
                profile_sigmas_sq = numpy.square(profile_sigmas)
                profile_tof = calc_profile_by_zcode_parameters(
                    delta_t_2d, d_hkl_2d,
                    profile_sigmas_sq[0], profile_sigmas_sq[1], profile_sigmas_sq[2],
                    profile_gammas[0], profile_gammas[1], profile_gammas[2],
                    profile_rs[0], profile_rs[1], profile_rs[2],
                    profile_alphas[0], profile_alphas[1],
                    profile_betas[0], profile_betas[1], profile_betas[2])

            dict_in_out_phase["profile_tof"] = profile_tof

        # flags_p_scale
        iint_m_plus = iint_plus * multiplicity_hkl
        iint_m_minus = iint_minus * multiplicity_hkl

        dict_in_out_phase["iint_plus_with_factors"] = 0.5 * \
            p_scale * iint_m_plus*lorentz_factor*wavelength_4_hkl
        dict_in_out_phase["iint_minus_with_factors"] = 0.5 * \
            p_scale * iint_m_minus*lorentz_factor*wavelength_4_hkl
        if flag_texture:
            # 0.5 to have the same meaning for the scale factor as in FullProf
            signal_plus = 0.5 * p_scale * lorentz_factor * \
                (profile_tof * (iint_m_plus * wavelength_4_hkl *
                 preferred_orientation)[na, :]).sum(axis=1)  # sum over hkl
            signal_minus = 0.5 * p_scale * lorentz_factor * \
                (profile_tof * (iint_m_minus * wavelength_4_hkl *
                 preferred_orientation)[na, :]).sum(axis=1)
            dict_in_out_phase["iint_plus_with_factors"] *= preferred_orientation
            dict_in_out_phase["iint_minus_with_factors"] *= preferred_orientation
        else:
            signal_plus = 0.5 * p_scale * lorentz_factor * \
                (profile_tof * (iint_m_plus * wavelength_4_hkl)
                 [na, :]).sum(axis=1)
            signal_minus = 0.5 * p_scale * lorentz_factor * \
                (profile_tof * (iint_m_minus *
                 wavelength_4_hkl)[na, :]).sum(axis=1)

        dict_in_out_phase["signal_plus"] = signal_plus
        dict_in_out_phase["signal_minus"] = signal_minus
        total_signal_plus += signal_plus
        total_signal_minus += signal_minus

    total_signal_plus *= spectrum_incident
    total_signal_minus *= spectrum_incident
    if flag_dict:
        dict_in_out["signal_plus"] = total_signal_plus
        dict_in_out["signal_minus"] = total_signal_minus

    if ("signal_exp_plus" in dict_tof_keys) and ("signal_exp_minus" in dict_tof_keys):
        signal_exp_plus = dict_tof["signal_exp_plus"]
        signal_exp_minus = dict_tof["signal_exp_minus"]
        if flag_dict:
            dict_in_out["signal_exp_plus"] = signal_exp_plus
            dict_in_out["signal_exp_minus"] = signal_exp_minus
        flag_chi_sq_sum, flag_chi_sq_difference = True, True

        if "flag_chi_sq_sum" in dict_tof_keys:
            flag_chi_sq_sum = dict_tof["flag_chi_sq_sum"]

        if "flag_chi_sq_difference" in dict_tof_keys:
            flag_chi_sq_difference = dict_tof["flag_chi_sq_difference"]

        if flag_chi_sq_sum:
            signal_exp = signal_exp_plus[0, :] + signal_exp_minus[0, :]
            signal_sigma = numpy.sqrt(numpy.square(
                signal_exp_plus[1, :]) + numpy.square(signal_exp_minus[1, :]))

    else:
        signal_exp = dict_tof["signal_exp"][0, :]
        signal_sigma = dict_tof["signal_exp"][1, :]
        if flag_dict:
            dict_in_out["signal_exp"] = dict_tof["signal_exp"]
        flag_chi_sq_sum = True
        flag_chi_sq_difference = False

    chi_sq = 0.
    n_point = 0
    if flag_chi_sq_sum:
        in_points = numpy.logical_not(excluded_points)
        total_signal_sum = total_signal_plus + total_signal_minus + signal_background
        chi_sq_sum = ((numpy.square(
            (signal_exp - total_signal_sum)/signal_sigma)*in_points)).sum(axis=0)
        chi_sq += chi_sq_sum
        n_point += numpy.sum(in_points)

    if flag_chi_sq_difference:
        signal_exp_diff = signal_exp_plus[0, :] - signal_exp_minus[0, :]
        signal_sigma_diff = numpy.sqrt(numpy.square(
            signal_exp_plus[1, :]) + numpy.square(signal_exp_minus[1, :]))
        total_signal_diff = total_signal_plus - total_signal_minus
        chi_sq_diff = (numpy.square(
            (signal_exp_diff - total_signal_diff)/signal_sigma_diff)).sum(axis=0)
        chi_sq += chi_sq_diff
        n_point += signal_exp_diff.shape[0]
    if numpy.isnan(chi_sq):
        chi_sq = 1e30

    flags_pd = get_flags(dict_tof)
    l_flags_crystal = [get_flags(dict_crystal)
                       for dict_crystal in dict_crystals]

    l_parameter_name = []
    for way, flags in flags_pd.items():
        pd_type_name = dict_tof["type_name"]
        ind_1d = numpy.atleast_1d(numpy.argwhere(flags))  # .flatten()
        parameter_name = [(pd_type_name, ) + way + (tuple(ind_1d[ind, :]), )
                          for ind in range(ind_1d.shape[0])]
        l_parameter_name.extend(parameter_name)

    for flags_crystal, dict_crystal in zip(l_flags_crystal, dict_crystals):
        for way, flags in flags_crystal.items():
            crystal_type_name = dict_crystal["type_name"]
            ind_1d = numpy.atleast_1d(numpy.argwhere(flags))  # .flatten()
            parameter_name = [(crystal_type_name, ) + way + (tuple(ind_1d[ind, :]), )
                              for ind in range(ind_1d.shape[0])]
            l_parameter_name.extend(parameter_name)

    # if flag_background_coefficients:
    #     pass

    der_chi_sq = numpy.zeros((len(l_parameter_name), ), dtype=float)
    dder_chi_sq = numpy.zeros(
        (len(l_parameter_name), len(l_parameter_name)), dtype=float)

    return chi_sq, n_point, der_chi_sq, dder_chi_sq, l_parameter_name



from typing import NoReturn
import numpy

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class DiffrnRadiation(ItemN):
    """
    Polarization of the neutron beam.

    Attributes:
        - polarization  (float)
        - efficiency (float)

    """
    ATTR_MANDATORY_NAMES = ("polarization", "efficiency")
    ATTR_MANDATORY_TYPES = (float, float)
    ATTR_MANDATORY_CIF = ("polarization", "efficiency")

    ATTR_OPTIONAL_NAMES = ()
    ATTR_OPTIONAL_TYPES = ()
    ATTR_OPTIONAL_CIF = ()

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("polarization", "efficiency")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {"polarization": 1., "efficiency": 1.}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "diffrn_radiation"

    def __init__(self, **kwargs) -> NoReturn:
        super(DiffrnRadiation, self).__init__()

        # defined for any integer and float parameters
        # D_MIN = {"polarization": -1., "efficiency": -1.}
        D_MIN = {}

        # defined for ani integer and float parameters
        # D_MAX = {"polarization": 1., "efficiency": 1.}
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def get_dictionary(self):
        res = {}
        beam_polarization = self.polarization
        flipper_efficiency = self.efficiency
        res["beam_polarization"] = numpy.array([beam_polarization], dtype=float)
        res["flipper_efficiency"] = numpy.array([flipper_efficiency], dtype=float)
        res["flags_beam_polarization"] = numpy.array([self.polarization_refinement], dtype=bool)
        res["flags_flipper_efficiency"] = numpy.array([self.efficiency_refinement], dtype=bool)
        return res


class DiffrnRadiationL(LoopN):
    """
    Description of beam polarization in loop.

    """
    ITEM_CLASS = DiffrnRadiation
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(DiffrnRadiationL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

# s_cont = """
# loop_
#  _diffrn_radiation_polarization
#  _diffrn_radiation_efficiency
#  1. 1.
#  1. -0.87()
#  5() -7
# """

# obj = DiffrnRadiationL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj.get_variable_names(), end="\n\n")



"""
Functions:
    - gammanu_to_tthphi
    - tthphi_to_gammanu
    - recal_int_to_tthphi_grid
    - recal_int_to_gammanu_grid
    - app_grid
    - app_2d

"""
import math


def gammanu_to_tthphi(gamma, nu):
    """Transfer gamma-nu to ttheta-phi.

    gamma
        is the angle in the equatorial plane

    nu
        is the angle between equatorial plane and scattering neutron
        [-pi/2, pi/2]

    return
        ttheta is diffraction angle
        phi is polar angle
    """
    ttheta = math.acos(math.cos(gamma)*math.cos(nu))
    phi = math.atan2(math.sin(nu), math.cos(nu)*math.sin(gamma))
    return ttheta, phi


def tthphi_to_gammanu(tth, phi):
    """Transfer ttheta-phi to gamma-nu."""
    gamma = math.atan2(math.cos(phi)*math.sin(tth), math.cos(tth))
    nu = math.asin(math.sin(tth)*math.sin(phi))
    return gamma, nu


def recal_int_to_tthphi_grid(l_gamma_grid, l_nu_grid, ll_int_grid,
                             l_ttheta_grid, l_phi_grid):
    l_point = []
    for phi in l_phi_grid:
        for tth in l_ttheta_grid:
            point = tthphi_to_gammanu(tth, phi)
            l_point.append(point)

    l_int_dang_grid = app_grid(ll_int_grid, l_gamma_grid, l_nu_grid, l_point)

    n_tth = len(l_ttheta_grid)
    lint_out = [[l_int_dang_grid[ind_tth+ind_phi*n_tth]
                 for ind_tth, tth in enumerate(l_ttheta_grid)]
                for ind_phi, phi in enumerate(l_phi_grid)]
    return lint_out


def recal_int_to_gammanu_grid(ltth_grid, l_phi_grid, ll_int_grid,
                              l_gamma_grid, l_nu_grid):
    l_point = []
    for nu in l_nu_grid:
        for gamma in l_gamma_grid:
            point = gammanu_to_tthphi(gamma, nu)
            l_point.append(point)

    l_int_dang_grid = app_grid(ll_int_grid, ltth_grid, l_phi_grid, l_point)

    n_gamma = len(l_gamma_grid)
    lint_out = [[l_int_dang_grid[ind_gamma+ind_nu*n_gamma]
                 for ind_gamma, gamma in enumerate(l_gamma_grid)]
                for ind_nu, nu in enumerate(l_nu_grid)]
    return lint_out


def app_grid(mat_xy, x_grid, y_grid, l_point):
    l_res = []
    min_x, max_x = float(x_grid[0]), float(x_grid[-1])
    min_y, max_y = float(y_grid[0]), float(y_grid[-1])
    n_x, n_y = len(x_grid), len(y_grid)
    step_x = (max_x - min_x)/float(n_x-1)
    step_y = (max_y - min_y)/float(n_y-1)
    for ipoint, point in enumerate(l_point):
        try:
            val_x = point[0]
            val_y = point[1]
        except Exception:
            return []

        n_tot_x = (val_x - min_x)/step_x
        n_tot_y = (val_y - min_y)/step_y
        if all([(n_tot_x >= 0.), (n_tot_x < float(n_x-1)),
                (n_tot_y >= 0.), (n_tot_y < float(n_y-1))]):
            nx = n_tot_x % 1.
            ny = n_tot_y % 1.
            ind_x = int(n_tot_x//1.)
            ind_y = int(n_tot_y//1.)
            res = app_2d(mat_xy[ind_y][ind_x], mat_xy[ind_y][ind_x+1],
                         mat_xy[ind_y+1][ind_x], mat_xy[ind_y+1][ind_x+1],
                         nx, ny)
        else:
            res = None
        l_res.append(res)
    return l_res


def app_2d(f11, f12, f21, f22, nx, ny):
    try:
        res2 = f21 - nx*(f21-f22)
        res1 = f11 - nx*(f11-f12)
        res = res1 - ny*(res1-res2)
    except Exception:
        res = None
    return res



from typing import NoReturn
import math
import numpy
import matplotlib.pyplot as plt

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class DiffrnRefln(ItemN):
    """
    The flip ratios measured in the single diffraction experiment.

    Data items in the DIFFRN_REFLN category record details about
    the intensities measured in the diffraction experiment.

    The DIFFRN_REFLN data items refer to individual intensity
    measurements and must be included in looped lists.

    Attributes:
        - 

    """
    ATTR_MANDATORY_NAMES = ("index_h", "index_k", "index_l")
    ATTR_MANDATORY_TYPES = (int, int, int)
    ATTR_MANDATORY_CIF = ("index_h", "index_k", "index_l")

    ATTR_OPTIONAL_NAMES = ("fr", "fr_sigma", "fr_calc", "intensity", "intensity_sigma", "intensity_calc",
                           "excluded", "sintlambda")
    ATTR_OPTIONAL_TYPES = (float, float, float, float, float, float, bool, float)
    ATTR_OPTIONAL_CIF = ("fr", "fr_sigma", "fr_calc", "intensity", "intensity_sigma", "intensity_calc",
                         "excluded", "sint/lambda")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"fr_calc": "{:.5f}", "intensity_calc": "{:.2f}",
                 "sintlambda": "{:.5f}",
                 "intensity": "{:.2f}", "intensity_sigma": "{:.2f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {"excluded": False}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "diffrn_refln"

    def __init__(self, **kwargs) -> NoReturn:
        super(DiffrnRefln, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"fr": 0., "fr_sigma": 0., "fr_calc": 0.,
                 "intensity_calc": 0.,
                 "sintlambda": 0.,
                 "intensity": 0., "intensity_sigma":0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class DiffrnReflnL(LoopN):
    """
    Flip ratios measured in the single diffraction experiment.

    """
    ITEM_CLASS = DiffrnRefln
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(DiffrnReflnL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def report(self):
        return self.report_agreement_factor_exp() + "\n" + self.report_chi_sq_exp()

    def report_agreement_factor_exp(self):
        """
        Make a report about experimental agreement factor in string format.
        """

        l_chi_sq_exp, l_ag_f_exp = [], []
        l_diff = []
        n_3s, n_2s, n_1s, n_0s = 0, 0, 0, 0
        l_hkl = [(int(_1), int(_2), int(_3)) for _1, _2, _3 in
                 zip(self.index_h, self.index_k, self.index_l)]
        for _hkl, _fr_1, _fr_sigma_1 in zip(l_hkl, self.fr, self.fr_sigma):
            _diff = abs(_fr_1-1.)/float(_fr_sigma_1)
            if _diff >= 3.:
                n_3s += 1
            elif _diff >= 2.:
                n_2s += 1
            elif _diff >= 1.:
                n_1s += 1
            else:
                n_0s += 1
            if _fr_1 <= 1.:
                if math.isclose(_fr_1, 0.):
                    l_diff.append(1./(_fr_1+_fr_sigma_1)-1.)
                else:
                    l_diff.append(1./_fr_1-1.)
            else:
                l_diff.append(_fr_1-1.)

            _mhkl = (-1 * _hkl[0], -1 * _hkl[1], -1 * _hkl[2])
            if _mhkl in l_hkl:
                ind_mhkl = l_hkl.index(_mhkl)
                _fr_2, _fr_sigma_2 = self.fr[ind_mhkl], self.fr_sigma[ind_mhkl]
                _fr_sigma = 1. / (_fr_sigma_1 ** (-2) + _fr_sigma_2 ** (-2)) ** 0.5
                _fr_average = (_fr_1 * _fr_sigma_1 ** (-2) + _fr_2 * _fr_sigma_2 ** (-2)) * _fr_sigma ** 2
                delta_fr = abs(_fr_1 - _fr_average)
                chi_sq_exp = (delta_fr / _fr_sigma_1) ** 2
                l_chi_sq_exp.append(chi_sq_exp)
                if math.isclose(_fr_1-1., 0):
                    ag_f_exp = abs((_fr_1 - _fr_average) / (_fr_1 - 1.+_fr_sigma_1))
                else:
                    ag_f_exp = abs((_fr_1 - _fr_average) / (_fr_1 - 1.))
                l_ag_f_exp.append(ag_f_exp)
                # print("hkl: {:4} {:4} {:4}".format(_hkl[0], _hkl[1], _hkl[2]))
                # print("chi_sq_exp: {:.3f} ".format(chi_sq_exp))
                # print("ag_f_exp: {:.3f} ".format(ag_f_exp))
        ls_out = ["# Experimental agreement factor\n"]
        n = n_0s + n_1s + n_2s + n_3s
        ls_out.append(f"\n| Number of measured reflections:| {n:4}|")
        ls_out.append(f"  |    range of h is |{min([_[0] for _ in l_hkl]):3},{max([_[0] for _ in l_hkl]):3} |")
        ls_out.append(f"  |             k is |{min([_[1] for _ in l_hkl]):3},{max([_[1] for _ in l_hkl]):3} |")
        ls_out.append(f"  |             l is |{min([_[2] for _ in l_hkl]):3},{max([_[2] for _ in l_hkl]):3} |")
        ls_out.append(f" |max(FR_exp - 1) is |{max(l_diff):5.3f} |")
        ls_out.append("\n N+1 > abs(FR_exp - 1)/FR_sigma > N: ")
        ls_out.append(f" |abs(FR_exp - 1)/FR_sigma < 1: |{n_0s:4}| {100*float(n_0s)/float(n):5.1f}% | ")
        ls_out.append(f" |                    N = 1: |{n_1s:4}| {100*float(n_1s)/float(n):5.1f}%  |")
        ls_out.append(f" |                    N = 2: |{n_2s:4}| {100 * float(n_2s) / float(n):5.1f}% |")
        ls_out.append(f" |abs(FR_exp - 1)/FR_sigma > 3: |{n_3s:4}| {100 * float(n_3s) / float(n):5.1f}% |")

        n_friedel = len(l_chi_sq_exp)
        ls_out.append(f"\n|Total number of Friedel reflections is |{n_friedel:}.|")
        if n_friedel != 0:
            ls_out.append(f"|  (abs(FR_exp-FR_av.)/FR_sigma)^2  is| {sum(l_chi_sq_exp) / n_friedel:.2f}|")
            ls_out.append(f"|   abs(FR_exp-FR_av.)/abs(FR_exp-1) per reflection is |{(100*sum(l_ag_f_exp)/n_friedel):.2f}% |")

        return "\n".join(ls_out)

    def calc_chi_sq_points(self) -> (float, int):
        """Calculate chi_sq and points if fr, fr_sigma, fr_calc are given."""
        if (self.is_attribute("fr") & self.is_attribute("fr_sigma") &
                self.is_attribute("fr_calc")):

            np_excl = numpy.array(self.excluded, dtype=bool)
            if numpy.all(np_excl):
                return (0., 0)

            flag_in = numpy.logical_not(np_excl)
            np_fr = numpy.array(self.fr, dtype=float)[flag_in]
            np_fr_calc = numpy.array(self.fr_calc, dtype=float)[flag_in]
            np_fr_sigma = numpy.array(self.fr_sigma, dtype=float)[flag_in]

            chi_sq = numpy.sum(numpy.square((np_fr-np_fr_calc)/np_fr_sigma))
            points = np_fr.size
            return chi_sq, points

        return (0., 0)


    def report_chi_sq_exp(self, cell=None) -> str:
        """
        Make a report about experimental chi_sq in string format.

        cell is unit cell object.
        """
        ls_out = []
        l_hkl = [(_1, _2, _3) for _1, _2, _3 in zip(self.index_h, self.index_k, self.index_l)]
        l_fr, l_fr_sigma, l_fr_calc = self.fr, self.fr_sigma, self.fr_calc
        if (l_fr is None) & (l_fr_sigma is None) & (l_fr_calc is None):
            return "\n".join(ls_out)
        n = len(l_fr)
        n_1s, n_2s, n_3s = 0, 0, 0
        l_chi_sq, l_worsest, l_af_f, l_af_r = [], [], [], []
        for _hkl, _fr, _fr_sigma, _fr_calc in zip(l_hkl, l_fr, l_fr_sigma, l_fr_calc):
            _diff = abs(float(_fr-_fr_calc)/float(_fr_sigma))
            l_chi_sq.append(_diff**2)
            if math.isclose(_fr, 0.):
                l_af_f.append(abs(float(_fr - _fr_calc) / float(_fr_sigma)))
            else:
                l_af_f.append(abs(float(_fr - _fr_calc) / float(_fr)))
            if not(math.isclose(_fr, 1.)):
                l_af_r.append(abs(float(_fr-_fr_calc)/float(_fr-1.)))
            if _diff <= 1.:
                n_1s +=1
            elif _diff <= 2.:
                n_2s += 1
            elif _diff <= 3.:
                n_3s += 1
            else:
                l_worsest.append((_hkl, _fr, _fr_sigma, _fr_calc, _diff))
        ls_out.append("# Chi_sq experimental")
        ls_out.append(f"Total number of reflections is {n:}")
        ls_out.append(f"|  (abs(FR_exp-FR_mod)/FR_sigma)^2 per reflection is |{sum(l_chi_sq)/float(n):.2f}|")
        ls_out.append(f"|   abs(FR_exp-FR_mod)/FR_exp      per reflection is |{100*sum(l_af_f) / float(n):.2f}%|")
        ls_out.append(f"|   abs(FR_exp-FR_mod)/abs(FR_exp-1)    per reflection is |{100*sum(l_af_r) / float(n):.2f}%|")
        ls_out.append("           (reflections with FR_exp = 1 are excluded)")
        n_worsest = len(l_worsest)
        ls_out.append("\n## Reflections in range  ")
        ls_out.append(" (N-1)*FR_sigma < abs(FR_exp - FR_mod) < N*FR_sigma: ")
        ls_out.append(f"|      N = 1:| {n_1s:}/{n:} ={100*float(n_1s)/float(n):5.1f}% ({2*34.1:4.1f}%, three sigma rule) |")
        ls_out.append(f"|      N = 2:| {n_2s:}/{n:} ={100 * float(n_2s) / float(n):5.1f}% ({2*(13.6):4.1f}%, three sigma rule)|")
        ls_out.append(f"|      N = 3:| {n_3s:}/{n:} ={100 * float(n_3s) / float(n):5.1f}% ({2*(2.1):4.1f}%, three sigma rule)|")
        ls_out.append(f"|      N > 3:| {n_worsest:}/{n:} ={100 * float(n_worsest) / float(n):5.1f}% ({2*(0.1):4.1f}%, three sigma rule)|")
        l_worsest.sort(key=lambda x: x[4], reverse=True)
        if len(l_worsest) > 1:
            if n_worsest > 10: n_worsest = 10
            ls_out.append("\n## The ten worsest reflections:")
            ls_out.append("|  h | k | l  |     FR |FR_sigma | FR_calc | diff|")
            for (_hkl, _fr, _fr_sigma, _fr_calc, _diff) in l_worsest[:n_worsest]:
                ls_out.append(f"|{_hkl[0]:3}|{_hkl[1]:3}|{_hkl[2]:3}|{_fr:9.5f}|{_fr_sigma:9.5f}|{_fr_calc:9.5f}|{_diff:6.1f}|")
        if cell is not None:
            np_h = numpy.array(self.index_h, dtype=int)
            np_k = numpy.array(self.index_k, dtype=int)
            np_l = numpy.array(self.index_l, dtype=int)
            np_sthovl = cell.calc_sthovl(index_h=np_h, index_k=np_k,
                                         index_l=np_l)
            np_fr, np_fr_sigma = numpy.array(l_fr, dtype=float), numpy.array(l_fr_sigma, dtype=float) 
            np_fr_calc = numpy.array(l_fr_calc, dtype=float)
            np_diff = numpy.abs((np_fr-np_fr_calc)/np_fr_sigma)
            n_bins = 10
            np_val, np_sthovl_bins = numpy.histogram(np_sthovl, bins=n_bins)
            ls_out.append("\n## Distribution of reflection in sin(theta)/lambda range")
            ls_out.append("|sthovl_1 |sthovl_2 | (Exp-Mod)/Sigma |n_points|")
            for _1, _2, _3 in zip(np_sthovl_bins[:-1], np_sthovl_bins[1:], np_val):
                np_flag = numpy.logical_and(np_sthovl >= _1, np_sthovl < _2) 
                res = np_diff[np_flag]
                if res.size == 0:
                    ls_out.append(f"|{_1:8.3f} |{_2:8.3f}|       None    |      0|")
                else:
                    ls_out.append(f"|{_1:8.3f} |{_2:8.3f}|    {np_diff[np_flag].mean():7.3f} |{_3:10}|")
            numpy.histogram(np_sthovl, bins=n_bins)
        return "\n".join(ls_out)

    def plots(self):
        return [self.plot_fr_vs_fr_calc(),
                self.plot_asymmetry_vs_asymmetry_calc(),
                self.plot_intensity_vs_intensity_calc(),]
    
    def plot_fr_vs_fr_calc(self):
        """Plot experimental fr vs. fr_calc
        """
        if not(self.is_attribute("fr") & self.is_attribute("fr_sigma") &
               self.is_attribute("fr_calc")):
            return 

        fig, ax = plt.subplots()
        np_excl = numpy.array(self.excluded, dtype=bool)
        flag_in = numpy.logical_not(np_excl)
        if numpy.all(np_excl):
            ax.set_title("Flip Ratio: I_plus / I_minus")
            fr_min = 0.
            fr_max = 10.
        else:
            np_fr = numpy.array(self.fr, dtype=float)[flag_in]
            np_fr_calc = numpy.array(self.fr_calc, dtype=float)[flag_in]
            np_fr_sigma = numpy.array(self.fr_sigma, dtype=float)[flag_in]

            chi_sq_per_n = numpy.square((np_fr-np_fr_calc)/np_fr_sigma).sum()/np_fr.shape[0]
            ax.set_title(r"Flip Ratio: $I_{plus} / I_{minus}$, $\chi^2/n=$" + f"{chi_sq_per_n:.2f}.")
    
            np_fr_1 = np_fr - np_fr_sigma
            np_fr_2 = np_fr + np_fr_sigma
    
            fr_min = min([min(np_fr_1), min(self.fr_calc)])
            fr_max = max([max(np_fr_2), max(self.fr_calc)])
            ax.plot([fr_min, fr_max], [fr_min, fr_max], "k:")
            ax.errorbar(np_fr_calc, np_fr, yerr=np_fr_sigma, fmt="ko", alpha=0.2)

        flag_excl = numpy.logical_not(flag_in)
        np_fr_excl = numpy.array(self.fr, dtype=float)[flag_excl]
        np_fr_calc_excl = numpy.array(self.fr_calc, dtype=float)[flag_excl]
        np_fr_sigma_excl = numpy.array(self.fr_sigma, dtype=float)[flag_excl]
        ax.errorbar(np_fr_calc_excl, np_fr_excl, yerr=np_fr_sigma_excl,
                    fmt="rs", alpha=0.2, label="excluded")

        ax.set_xlim(fr_min, fr_max)
        ax.set_ylim(fr_min, fr_max)
        
        ax.set_xlabel("Flip ratio (model)")
        ax.set_ylabel('Flip ratio (experiment)')
        ax.set_aspect(1)
        fig.tight_layout()
        return (fig, ax)

    def plot_asymmetry_vs_asymmetry_calc(self):
        """Plot experimental fr vs. fr_calc
        """
        if not(self.is_attribute("fr") & self.is_attribute("fr_sigma") &
               self.is_attribute("fr_calc")):
            return 

        fig, ax = plt.subplots()
        np_excl = numpy.array(self.excluded, dtype=bool)
        flag_in = numpy.logical_not(np_excl)
        np_fr = numpy.array(self.fr, dtype=float)[flag_in]
        np_fr_calc = numpy.array(self.fr_calc, dtype=float)[flag_in]
        np_fr_sigma = numpy.array(self.fr_sigma, dtype=float)[flag_in]

        asymmetry = (np_fr - 1.)/(np_fr + 1.)
        asymmetry_sigma = np_fr_sigma*numpy.sqrt(2.)*numpy.sqrt(numpy.square(np_fr)+1.) / \
            numpy.square(np_fr+1.)
        asymmetry_calc = (np_fr_calc - 1.)/(np_fr_calc + 1.)
        chi_sq_per_n = numpy.square((asymmetry-asymmetry_calc)/asymmetry_sigma).sum()/asymmetry.size

        ax.set_title(r"Asymmetry parameter: $\frac{I_{plus}-I_{minus}}{I_{plus}+I_{minus}}$, $\chi^2/n=$" + f"{chi_sq_per_n:.2f}.")

        ax.plot([-1, 1], [-1, 1], "k:")
        ax.errorbar(asymmetry_calc, asymmetry, yerr=asymmetry_sigma, fmt="ko",
                    alpha=0.2)
        ax.set_xlim(-1, 1)
        ax.set_ylim(-1, 1)

        ax.set_xlabel("Asymmetry (model)")
        ax.set_ylabel('Asymmetry (experiment)')
        ax.set_aspect(1)
        fig.tight_layout()
        return (fig, ax)

    def plot_intensity_vs_intensity_calc(self):
        """Plot experimental intensity vs. intensity_calc
        """
        if not(self.is_attribute("intensity") & self.is_attribute("intensity_sigma") &
               self.is_attribute("intensity_calc")):
            return 

        fig, ax = plt.subplots()
        np_excl = numpy.array(self.excluded, dtype=bool)
        flag_in = numpy.logical_not(np_excl)
        if numpy.all(np_excl):
            ax.set_title("Intensity")
            intensity_min = 0.
            intensity_max = 10.
        else:
            np_intensity = numpy.array(self.intensity, dtype=float)[flag_in]
            np_intensity_calc = numpy.array(self.intensity_calc, dtype=float)[flag_in]
            np_intensity_sigma = numpy.array(self.intensity_sigma, dtype=float)[flag_in]

            chi_sq_per_n = numpy.square((np_intensity-np_intensity_calc)/np_intensity_sigma).sum()/np_intensity.shape[0]
            agreement_factor = numpy.abs(np_intensity-np_intensity_calc).sum()/numpy.abs(np_intensity).sum() * 100.
            ax.set_title(r"Intensity, $\chi^2/n=$" + f"{chi_sq_per_n:.2f} "+ r" $RF^2=$"+f"{agreement_factor:.1f}%")
    
            np_intensity_1 = np_intensity - np_intensity_sigma
            np_intensity_2 = np_intensity + np_intensity_sigma
    
            intensity_min = min([min(np_intensity_1), min(self.intensity_calc)])
            intensity_max = max([max(np_intensity_2), max(self.intensity_calc)])
            ax.plot([intensity_min, intensity_max], [intensity_min, intensity_max], "k:")
            ax.errorbar(np_intensity_calc, np_intensity, yerr=np_intensity_sigma, fmt="ko", alpha=0.2)

        flag_excl = numpy.logical_not(flag_in)
        np_intensity_excl = numpy.array(self.intensity, dtype=float)[flag_excl]
        np_intensity_calc_excl = numpy.array(self.intensity_calc, dtype=float)[flag_excl]
        np_intensity_sigma_excl = numpy.array(self.intensity_sigma, dtype=float)[flag_excl]
        ax.errorbar(np_intensity_calc_excl, np_intensity_excl, yerr=np_intensity_sigma_excl,
                    fmt="rs", alpha=0.2, label="excluded")

        ax.set_xlim(intensity_min, intensity_max)
        ax.set_ylim(intensity_min, intensity_max)
        
        ax.set_xlabel("Intensity (model)")
        ax.set_ylabel('Intensity (experiment)')
        ax.set_aspect(1)
        fig.tight_layout()
        return (fig, ax)


    def include_all_points(self):
        for item in self.items:
            item.excluded = False
        name = "numpy_excluded"
        if name in self.__dict__.keys():
            del self.__dict__[name]

    def exclude_by_fr_max(self, fr_max: float = 3.):
        """Exclude points by rules
        1./fr_max <= FR <= fr_max
        """
        if fr_max >= 1.:
            fr_1 = 1./fr_max
            fr_2 = fr_max
        else:
            fr_1 = fr_max
            fr_2 = 1./fr_max

        for item in self.items:
            if ((item.fr < fr_1) | (item.fr > fr_2)):
                item.excluded = True

        name = "numpy_excluded"
        if name in self.__dict__.keys():
            del self.__dict__[name]

    def get_dictionary(self):
        res = {}
        index_hkl = numpy.array([
            self.index_h, self.index_k,
            self.index_l], dtype=float)
        refln_fr_excl = numpy.array(self.excluded, dtype=bool)
        res["index_hkl"] = index_hkl
        res["flip_ratio_excluded"] = refln_fr_excl
        try:
            refln_fr_es = numpy.array([
                self.fr, self.fr_sigma], dtype=float)
            res["flip_ratio_es"] = refln_fr_es
        except AttributeError:
            refln_intensity_es = numpy.array([
                self.intensity, self.intensity_sigma], dtype=float)
            res["intensity_es"] = refln_intensity_es
        return res

# s_cont = """
#   loop_
#   _diffrn_refln_index_h
#   _diffrn_refln_index_k
#   _diffrn_refln_index_l
#   _diffrn_refln_fr
#   _diffrn_refln_fr_sigma
#       0    0    8   0.64545   0.01329 
#       2    0    6   1.75682   0.0454
# """

# obj = DiffrnReflnL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj.report_agreement_factor_exp(), end="\n\n")
# print(obj.numpy_fr_sigma, end="\n\n")



"""Transformation of markdown in different formals
"""


def md_to_html(s_content: str):
    if s_content.strip() == "":
        return ""
    l_html = []
    l_content = s_content.split("\n")
    f_table_head = False
    f_ul = False
    f_ol = False

    for line in l_content:
        s_line = line.strip()
        if s_line.find("**") != -1:
            ls_line = []
            for i_val, val in enumerate(s_line.split("**")):
                if i_val%2 == 0:
                    ls_line.append(f"{val:}<b>")
                elif i_val%2 == 1:
                    ls_line.append(f"</b>{val:}")
            if i_val%2 == 0:
                ls_line.append(f"</b>")
            s_line = "".join(ls_line)

        if s_line.find("*") != -1:
            ls_line = []
            for i_val, val in enumerate(s_line.split("*")):
                if i_val%2 == 0:
                    ls_line.append(f"{val:}<i>")
                elif i_val%2 == 1:
                    ls_line.append(f"</i>{val:}")
            if i_val%2 == 0:
                ls_line.append(f"</i>")
            s_line = "".join(ls_line)

        if s_line.startswith("# "):
            l_html.append(f"<h1>{s_line.strip('# '):}</h1>\n")
        elif s_line.startswith("## "):
            l_html.append(f"<h2>{s_line.strip('# '):}</h2>\n")
        elif s_line.startswith("### "):
            l_html.append(f"<h3>{s_line.strip('# '):}</h3>\n")
        elif s_line.startswith("#### "):
            l_html.append(f"<h4>{s_line.strip('# '):}</h4>\n")
        elif s_line.startswith("|"):
            if len(set(s_line) - set("|:- ")) == 0:
                f_table_head = True
            else:
                if not(f_table_head):
                    l_html.append("<table>")
                    l_td = [f"<th>{th.strip():}</th>" for th in s_line.split("|") if th != ""]
                    f_table_head = True
                else:
                    l_td = [f"<td>{th.strip():}</td>" for th in s_line.split("|") if th != ""]
                l_html.append(f"<tr align='right'>{''.join(l_td):}</tr>")
        elif (s_line.startswith("-") | s_line.startswith("+")):
            if not(f_ul):
                f_ul = True
                l_html.append("<ul>")
            l_html.append(f"<li>{s_line[1:].strip():}</li>")
        elif (s_line.startswith("*.")):
            if not(f_ol):
                f_ol = True
                l_html.append("<ol>")
            l_html.append(f"<li>{s_line[1:].strip():}</li>")
        elif (s_line.startswith("![")):
            l_h = s_line.split("](")
            s_alt = l_h[0].lstrip("![")
            s_src = l_h[1].rstrip(")")
            l_html.append(f"<p><img alt='{s_alt:}' src='{s_src:}'></p>")
        else:
            if f_table_head:
                f_table_head = False
                l_html.append("</table>")
            if f_ul:
                f_ul = False
                l_html.append("</ul>")
            if f_ol:
                f_ol = False
                l_html.append("</ol>")
            l_html.append(f"<p>{s_line:}</p>")

    return "".join(l_html)





"""
Functions to work with strings.
"""

from typing import Tuple, Union
import numpy
from fractions import Fraction


def ciftext_to_html(text: str) -> str:
    """Convert cif text to html format."""
    l_text_new = []
    for line in text.split("\n"):
        if line.strip().startswith("###"):
            line_new = f"<h3>{line:}</h3>"
        elif line.strip().startswith("##"):
            line_new = f"<h2>{line:}</h2>"
        elif line.strip().startswith("#"):
            line_new = f"<h1>{line:}</h1>"
        else:
            line_new = line
        l_text_new.append(line_new)
    text_html = "<br>".join(l_text_new)
    # text_html = "<b>"+text.replace("\n", "<br>")+"<\b>"
    return text_html


def find_prefix(*argv) -> str:
    """Find prefix
    
    Example 1 (trivial)
    -------------------
    
    loop_
    _atom_site_label
    _atom_site_fract_x
    _atom_site_fract_y
    
    Output: "_atom_site"

    Example 2 (not trivial)
    -----------------------
    loop_
        _geom_torsion
        _geom_torsion_atom_site_label_1
        _geom_torsion_atom_site_label_2
    
    Output: "_geom_torsion"
    (name for item "geom_torsion" is label)
    
    """
    c_string = common_string(*argv)
    ind_point = c_string.rfind(".")
    if ind_point != -1:
        prefix = c_string[:ind_point]
    elif ((len(argv) > 1) & (c_string in argv)):
        prefix = c_string
    else:
        ind_und = c_string.rfind("_")
        prefix = c_string[:ind_und]
    return prefix


def common_string(*argv) -> str:
    """Return common begining of input strings."""
    ls_out = []
    for args in zip(*argv):
        arg_1 = args[0]
        flag = all([arg == arg_1 for arg in args])
        if flag:
            ls_out.append(arg_1)
        else:
            break
    return "".join(ls_out)



def string_to_value_error_mark(string: str) -> Tuple[float, Union[float, None], str]:
    """
    Convert string to float and error.

    Parameters
    ----------
    string : str
        DESCRIPTION.

    Returns
    -------
    value : float
        Value.
    error : float
        Error.

    """
    value, error, mark = None, None, ""
    ind_1 = string.find("(")
    s_sigma = ""
    if string == ".":
        pass 
    elif ind_1 != -1:
        ind_2 = string.find(")")
        if ind_2 > ind_1:
            s_sigma = string[(ind_1+1):ind_2]
            if not(s_sigma.isdigit()):
                s_sigma = ""
        str_1 = string.split("(")[0]
        value = float(str_1)
        mark = string[ind_2+1:].strip()
        if s_sigma != "":
            s_h = "".join(["0" if _h.isdigit() else _h for _h in
                           str_1[:-len(s_sigma)]])
            error = abs(float(s_h+s_sigma))
        else:
            error = 0.
    else:
        try:
            value = float(string)
        except ValueError:
            value = None
    return value, error, mark

def value_error_mark_to_string(value: float, error: float, mark: str) -> str:
    """
    Convert value and error to string

    Parameters
    ----------
    value : float
        DESCRIPTION.
    error : float
        DESCRIPTION.

    Returns
    -------
    str
        DESCRIPTION.

    """
    if ((value is None) | (value is numpy.nan)):
        string = "."
    if not((error == 0.) | (error is None)):
        val_hh = numpy.log10(error)
        n_power = int(val_hh)
        if val_hh <= 0:
            val_1 = numpy.round(value, decimals = -1*int(n_power)+2)
            val_2 = numpy.round(error, decimals = -1*int(n_power)+2)
            i_val_11 = int(val_1)
            s_sign = ""
            if val_1 < 0.:
                s_sign = "-"
            s_val_12 = ("{:}".format(int(abs(val_1) % 1.*10**(-n_power+2)))
                        ).rjust(int(-n_power+2), "0")
            val_2 = int(val_2*10**(-n_power+2))
            string = f"{s_sign:}{abs(i_val_11):}.{s_val_12:}({int(val_2):}){mark.strip():}"
        else:
            val_1 = numpy.round(value)
            val_2 = numpy.round(error)
            string = "{:}({:}){:}".format(int(val_1), int(val_2),mark.strip())
    elif (error == 0.):
        string = "{:}(){:}".format(value, mark.strip())
    else:
        string = "{:}".format(value)
    return string

##FIXME to del
#def string_to_value_error(string: str) -> Tuple[float, Union[float, None]]:
#    """
#    Convert string to float and error.
#
#    Parameters
#    ----------
#    string : str
#        DESCRIPTION.
#
#    Returns
#    -------
#    value : float
#        Value.
#    error : float
#        Error.
#
#    """
#    value, error = None, None
#    ind_1 = string.find("(")
#    s_sigma = ""
#    if value == ".":
#       pass 
#    elif ind_1 != -1:
#        ind_2 = string.find(")")
#        if ind_2 > ind_1:
#            s_sigma = string[(ind_1+1):ind_2]
#            if not(s_sigma.isdigit()):
#                s_sigma = ""
#        str_1 = string.split("(")[0]
#        value = float(str_1)
#        if s_sigma != "":
#            s_h = "".join(["0" if _h.isdigit() else _h for _h in
#                           str_1[:-len(s_sigma)]])
#            error = abs(float(s_h+s_sigma))
#        else:
#            error = 0.
#    else:
#        try:
#            value = float(string)
#        except ValueError:
#            value = None
#    return value, error
#
# #FIXME to del
# def value_error_to_string(value: float, error: float) -> str:
#     """
#     Convert value and error to string
# 
#     Parameters
#     ----------
#     value : float
#         DESCRIPTION.
#     error : float
#         DESCRIPTION.
# 
#     Returns
#     -------
#     str
#         DESCRIPTION.
# 
#     """
#     if ((value is None) | (value is numpy.nan)):
#         string = "."
#     if not((error == 0.) | (error is None)):
#         val_hh = numpy.log10(error)
#         n_power = int(val_hh)
#         if val_hh <= 0:
#             val_1 = numpy.round(value, decimals = -1*int(n_power)+2)
#             val_2 = numpy.round(error, decimals = -1*int(n_power)+2)
#             i_val_11 = int(val_1)
#             s_sign = ""
#             if val_1 < 0.:
#                 s_sign = "-"
#             s_val_12 = ("{:}".format(int(abs(val_1) % 1.*10**(-n_power+2)))
#                         ).rjust(int(-n_power+2), "0")
#             val_2 = int(val_2*10**(-n_power+2))
#             string = f"{s_sign:}{abs(i_val_11):}.{s_val_12:}({int(val_2):})"
#         else:
#             val_1 = numpy.round(value)
#             val_2 = numpy.round(error)
#             string = "{:}({:})".format(int(val_1), int(val_2))
#     elif (error == 0.):
#         string = "{:}()".format(value)
#     else:
#         string = "{:}".format(value)
#     return string
# 

def transform_string_to_r_b(name: str, labels=("x", "y", "z")) -> Tuple:
    """
    transform string to rotation part and offset: 
    x,y,-z   ->   0.0 1 0 0  0.0 0 1 0  0.0 0 0 -1
    """
    l_name = "".join(name.strip().split()).lstrip("(").rstrip(")").split(",")
    rij, bi = [], []
    for _name in l_name:
        coefficients, offset = transform_string_to_digits(_name, labels)
        rij.append(coefficients)
        bi.append(offset)
    res_rij = numpy.array(rij, dtype=object)
    res_bi = numpy.array(bi, dtype=object)
    return res_rij, res_bi



def transform_string_to_digits(name: str, labels: Tuple[str]):
    """
    Multiplication has to be implicit, division must be explicit.
    White space within the string is optional 

    Example: 
    transform_string_to_digits('-a+2x/7-0.7t+4', ('a', 'x', 't')) = 
    = (Fraction(-1,1),fraction(2,7),Fraction(-7,10)), Fraction(4,1)
    """
    coefficients = []
    offset = Fraction(0, 1).limit_denominator(10)
    l_name = name.strip().replace(" ", "").replace("+", " +").replace("-", " -").split()
    for _label in labels:
        res = Fraction(0, 1).limit_denominator(10)
        for _name in l_name:
            if _label in _name:
                s_1 = _name.replace(_label, "").replace("+/", "+1/").replace("-/", "-1/")
                if s_1 == "": s_1 = "1"
                if s_1.startswith("/"): s_1 = "1" + s_1
                if s_1.endswith("+"): s_1 = s_1 + "1"
                if s_1.endswith("-"): s_1 = s_1 + "1"
                res += Fraction(s_1).limit_denominator(10)
        coefficients.append(res)
    res = Fraction(0, 1).limit_denominator(10)
    for _name in l_name:
        flag = all([not (_label in _name) for _label in labels])
        if flag:
            res += Fraction(_name).limit_denominator(10)
    offset = res
    return coefficients, offset


def transform_fraction_with_label_to_string(number: Fraction, label: str) -> str:
    if isinstance(number, Fraction):
        val = number
    else:
        val = Fraction(number).limit_denominator(10)
    if val == Fraction(0, 1):
        res = ""
    elif val == Fraction(1, 1):
        res = f"{label:}"
    elif val == Fraction(-1, 1):
        res = f"-{label:}"
    elif val.denominator == 1:
        res = f"{val.numerator}{label:}"
    elif val.numerator == 1:
        res = f"{label:}/{val.denominator}"
    else:
        res = f"{val.numerator}{label:}/{val.denominator}"
    return res


def transform_digits_to_string(labels: Tuple[str], coefficients, offset: Fraction) -> str:
    """
Form a string from digits.

Keyword arguments: 

    labels: the tuple of lablels (ex.: ('x', 'y', 'z') or ('a', 'b', 'c')))
    coefficients: the parameters in front of label (ex.: (1.0, 0.5, 0.0))
    offset: the number (ex.: 2/3)

Output arguments is string.

Example:
>>> transform_digits_to_string(('x', 'y', 'z'), (1.0, 0.5, 0.0), 0.6666667)
x+1/2y+2/3
    """
    l_res = []
    for _coefficient, _label in zip(coefficients, labels):
        _name = transform_fraction_with_label_to_string(_coefficient, _label)
        if _name == "":
            pass
        elif _name.startswith("-"):
            l_res.append(_name)
        elif l_res == []:
            l_res.append(_name)
        else:
            l_res.append(f"+{_name:}")
    _name = str(Fraction(offset).limit_denominator(10))
    if _name == "0":
        if l_res == []:
            l_res.append(_name)
    elif ((l_res == []) | (_name.startswith("-"))):
        l_res.append(_name)
    else:
        l_res.append(f"+{_name:}")
    return "".join(l_res)


def transform_r_b_to_string(r, b, labels=("x", "y", "z")) -> str:
    l_res = [transform_digits_to_string(labels, _ri, _bi) for _ri, _bi in zip(r, b)]
    return ",".join(l_res)



from typing import Union
import math
import numpy

from cryspy.A_functions_base.function_1_strings import \
    value_error_mark_to_string

import cryspy.A_functions_base.local_susceptibility as local_susceptibility

from cryspy.C_item_loop_classes.cl_1_cell import Cell
from cryspy.C_item_loop_classes.cl_1_atom_site_susceptibility import \
    AtomSiteSusceptibility, AtomSiteSusceptibilityL

def magnetization_ellipsoid_by_u_ij(
        cell: Cell,
        atom_site_susceptibility: AtomSiteSusceptibility):
    """Calc magnetization ellispoid through thermal parameters U_ij.
    """
    ucp = cell.get_unit_cell_parameters()
    
    q_rn = numpy.array([
        atom_site_susceptibility.chi_11, atom_site_susceptibility.chi_22, atom_site_susceptibility.chi_33,
        atom_site_susceptibility.chi_12, atom_site_susceptibility.chi_13, atom_site_susceptibility.chi_23], dtype=float)
    

    u_ij = local_susceptibility.calc_magnetization_ellipsoid_as_u(
        q_rn, ucp)[0]

    return u_ij


def report_main_axes_of_magnetization_ellipsoids(
        cell: Cell,
        atom_site_susceptibility: Union[AtomSiteSusceptibility, AtomSiteSusceptibilityL]):
    """Give a report about magnetizatin ellipsoid"""

    ls_out = []
    ucp = cell.get_unit_cell_parameters()
    a_s_m_a = atom_site_susceptibility

    if isinstance(atom_site_susceptibility, AtomSiteSusceptibilityL):
        a_s_s_items = atom_site_susceptibility.items
    elif isinstance(atom_site_susceptibility, AtomSiteSusceptibility):
        a_s_s_items = [atom_site_susceptibility, ]
    else:
        return ""

    for a_s_s in a_s_s_items:
        chi_xyz = numpy.array([a_s_s.chi_11, a_s_s.chi_22, a_s_s.chi_33,
        a_s_s.chi_12, a_s_s.chi_13, a_s_s.chi_23], dtype=float)

        moments, moments_sigma, eig_fields, eig_moments = \
            a_s_s.calc_main_axes_of_magnetization_ellipsoid(cell)
        
        u_ij = magnetization_ellipsoid_by_u_ij(cell, a_s_s)

        label = a_s_s.label 
        ls_out.append(f"For **`{label:}`** the principal axes of the atomic susceptibility tensor are:\n")
        # cycle over three main axes
        directions = eig_fields.transpose()

        ef = local_susceptibility.calc_ellipsoid_factor(chi_xyz, ucp)

        ls_out.append("|Axes  (mu_B/T)|Orientation: |X along inv.a| Y is [inv.a, c]|Z along c|")
        ls_out.append("|--------------|-------|----------|----------|----------|")
        for _val1, val_sigma, _direction in zip(
                moments, moments_sigma, directions):
            if math.isclose(val_sigma, 0.):
                s_val = f"{_val1: 9.5f}".rjust(9)
            else:
                s_param = value_error_mark_to_string(_val1, val_sigma, "")
                s_val = f"{s_param:}".rjust(9)
            ls_out.append(
                f"|  {s_val:} | along:| {_direction[0]: 9.5f}| \
{_direction[1]: 9.5f}| {_direction[2]: 9.5f}|")
        ls_out.append(f"\nEllipsoid factor {ef:.5f}.\n")

        ls_out.append("\nUse thermal parameters U_ij to plot ellispoid.\n")
        ls_out.append(f"|      U_11|      U_22|      U_33|      U_12|      \
U_13|      U_23|")
        ls_out.append(f"|----------|----------|----------|----------|------\
----|----------|")
        ls_out.append(
                f"|{u_ij[0]:9.3f}|{u_ij[1]:9.3f}|\
{u_ij[2]:9.3f}|{u_ij[3]:9.3f}|{u_ij[4]:9.3f}|\
{u_ij[5]:9.3f}|\n")
    # ls_out.append("Cartezian coordinate system is x||a*, z||c.")
    if len(ls_out) != 0:
        ls_out.insert(0, "# Magnetization ellipsoids\n")
    return "\n".join(ls_out)



from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class AtomLocalAxes(ItemN):
    """
    This category allows the definition of local axes around each
    atom in terms of vectors between neighbouring atoms.
    High-resolution X-ray diffraction methods enable the
    determination of the electron density distribution in crystal
    lattices and molecules, which in turn allows for a
    characterization of chemical interactions (Coppens, 1997;
    Koritsanszky & Coppens, 2001). This is accomplished by the
    construction of a mathematical model of the charge density
    in a crystal and then by fitting the parameters of such a
    model to the experimental pattern of diffracted X-rays. The
    model on which this dictionary is based is the so-called
    multipole formalism proposed by Hansen & Coppens (1978). In
    this model, the electron density in a crystal is described
    by a sum of aspherical "pseudoatoms" where the pseudoatom
    density has the form defined in the _atom_rho_multipole_* items.
    Each pseudoatom density consists of terms representing the
    core density, the spherical part of the valence density and
    the deviation of the valence density from sphericity. The
    continuous electron density in the crystal is then modelled
    as a sum of atom-centred charge distributions. Once the
    experimental electron density has been established, the
    "atoms in molecules" theory of Bader (1990) provides tools for
    the interpretation of the density distribution in terms of its
    topological properties.

    Mandatory attributes:
        - wavelength

    Optional attributes:
        - field
        - offset_ttheta
        - offset_phi

    """
    ATTR_MANDATORY_NAMES = ("atom_label", "atom0", "ax1", "atom1", "atom2",
                            "ax2")
    ATTR_MANDATORY_TYPES = (str, str, str, str, str, str)
    ATTR_MANDATORY_CIF = ("atom_label", "atom0", "ax1", "atom1", "atom2",
                            "ax2")

    ATTR_OPTIONAL_NAMES = ()
    ATTR_OPTIONAL_TYPES = ()
    ATTR_OPTIONAL_CIF = ()

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {"ax1": ["X", "Y", "Z", "-X", "-Y", "-Z"],
                     "ax2": ["X", "Y", "Z", "-X", "-Y", "-Z"]}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "atom_local_axes"

    def __init__(self, **kwargs) -> NoReturn:
        super(AtomLocalAxes, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class AtomLocalAxesL(LoopN):
    """
    This category allows the definition of local axes around each
    atom in terms of vectors between neighbouring atoms.
    High-resolution X-ray diffraction methods enable the
    determination of the electron density distribution in crystal
    lattices and molecules, which in turn allows for a
    characterization of chemical interactions (Coppens, 1997;
    Koritsanszky & Coppens, 2001). This is accomplished by the
    construction of a mathematical model of the charge density
    in a crystal and then by fitting the parameters of such a
    model to the experimental pattern of diffracted X-rays. The
    model on which this dictionary is based is the so-called
    multipole formalism proposed by Hansen & Coppens (1978). In
    this model, the electron density in a crystal is described
    by a sum of aspherical "pseudoatoms" where the pseudoatom
    density has the form defined in the _atom_rho_multipole_* items.
    Each pseudoatom density consists of terms representing the
    core density, the spherical part of the valence density and
    the deviation of the valence density from sphericity. The
    continuous electron density in the crystal is then modelled
    as a sum of atom-centred charge distributions. Once the
    experimental electron density has been established, the
    "atoms in molecules" theory of Bader (1990) provides tools for
    the interpretation of the density distribution in terms of its
    topological properties.

    """
    ITEM_CLASS = AtomLocalAxes
    ATTR_INDEX = "atom_label"
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(AtomLocalAxesL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

# s_cont = """
#   loop_
#   _atom_local_axes_atom_label
#   _atom_local_axes_atom0
#   _atom_local_axes_ax1
#   _atom_local_axes_atom1
#   _atom_local_axes_atom2
#   _atom_local_axes_ax2
#       Ni2+(1)  DUM0      Z    Ni2+(1)  N(1)      X
#   """

# obj = AtomLocalAxesL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["Ni2+(1)"], end="\n\n")



import numpy
import scipy
import scipy.interpolate

from cryspy.A_functions_base.matrix_operations import calc_m1_m2_m1t, calc_m_v

from cryspy.A_functions_base.unit_cell import \
    calc_sthovl_by_unit_cell_parameters, calc_matrix_t

from cryspy.A_functions_base.structure_factor import \
    calc_f_nucl_by_dictionary, \
    calc_f_charge_by_dictionary, \
    calc_sft_ccs_by_dictionary, \
    calc_index_hkl_multiplicity_in_range, \
    calc_f_m_perp_ordered_by_dictionary

from cryspy.A_functions_base.integrated_intensity_powder_diffraction import \
    calc_powder_iint_1d_para, calc_powder_iint_1d_ordered, calc_powder_iint_1d_mix

from cryspy.A_functions_base.preferred_orientation import calc_preferred_orientation_pd

from cryspy.A_functions_base.powder_diffraction_const_wavelength import \
    calc_profile_pseudo_voight, calc_lorentz_factor

from .rhochi_diffrn import get_flags


na = numpy.newaxis

def calc_background(ttheta, background_ttheta, background_intensity, flag_background_intensity: bool = False):
    x_p = numpy.copy(background_ttheta)
    y_p = numpy.copy(background_intensity)
    x_min = ttheta.min()
    x_max = ttheta.max()
    if x_p.min() > x_min:
        y_0 = (y_p[1]-y_p[0])*(x_min - x_p[0])/(x_p[1]-x_p[0]) + y_p[0]
        x_p = numpy.insert(x_p, 0, x_min)
        y_p = numpy.insert(y_p, 0, y_0)
    if x_p.max() <= x_max:
        x_max = x_max + 1.
        y_last = (y_p[-1]-y_p[-2])*(x_max - x_p[-2])/(x_p[-1]-x_p[-2]) + y_p[-2]
        x_p = numpy.append(x_p, x_max)
        y_p = numpy.append(y_p, y_last)
    x_left = x_p[:-1]
    x_right = x_p[1:]
    flags = numpy.logical_and(ttheta[:, na] >= x_left[na, :], ttheta[:, na] < x_right[na, :])
    p0 = numpy.argwhere(flags)[:,1]
    p1 = p0 + 1 
    intensity = (y_p[p1]-y_p[p0]) * (ttheta-x_p[p0])/(x_p[p1]-x_p[p0]) + y_p[p0]
    # f = scipy.interpolate.interp1d(
    #     background_ttheta, background_intensity, kind="linear", fill_value="extrapolate")
    # intensity = f(ttheta)
    dder = {}
    if flag_background_intensity:
        ttheta_shift = ttheta[:, na] - background_ttheta[na, :]
        diff_b_tth = background_ttheta[1:]-background_ttheta[:-1]
        # y_n + (y_np1-y_np)*(x-x_n)/(x_np1-x_n)
        # 1  -(x-x_n)/(x_np1-x_n) + (x-x_nm1)/(x_n-x_nm1) 
        dder_bkgr = numpy.zeros((ttheta.size, background_ttheta.size), dtype=float)
        dder["background_intensity"] = dder_bkgr 
    return intensity, dder 


def calc_chi_sq_for_pd_by_dictionary(
        dict_pd, dict_crystals, dict_in_out: dict = None, flag_use_precalculated_data: bool=False,
        flag_calc_analytical_derivatives: bool = False):
    """Calculate chi_sq for diffrn experiment.
    """
    if dict_in_out is None:
        flag_dict = False
        flag_use_precalculated_data = False
        dict_in_out_keys = []
    else:
        flag_dict = True
        dict_in_out_keys = dict_in_out.keys()
    dict_pd_keys = dict_pd.keys()

    phase_name = [hh["name"].lower() for hh in dict_crystals]

    excluded_points = dict_pd["excluded_points"]
    ttheta = dict_pd["ttheta"]
    offset_ttheta = dict_pd["offset_ttheta"]
    ttheta_zs = ttheta - offset_ttheta
    flags_offset_ttheta = dict_pd["flags_offset_ttheta"]
    if flag_dict:
        dict_in_out["ttheta"] = ttheta
        dict_in_out["ttheta_corrected"] = ttheta_zs  
        dict_in_out["excluded_points"] = excluded_points 


    wavelength = dict_pd["wavelength"]
    flags_wavelength = dict_pd["flags_wavelength"]
    radiation = dict_pd["radiation"]

    if "beam_polarization" in dict_pd_keys:
        beam_polarization = dict_pd["beam_polarization"]
        flipper_efficiency = dict_pd["flipper_efficiency"]
        magnetic_field = dict_pd["magnetic_field"]
        flags_beam_polarization = dict_pd["flags_beam_polarization"]
        flags_flipper_efficiency = dict_pd["flags_flipper_efficiency"]
    else:
        beam_polarization, flipper_efficiency, magnetic_field = 0., 0., 0.
        flags_beam_polarization, flags_flipper_efficiency = False, False

    sthovl_min = numpy.sin(0.5*ttheta_zs.min() - numpy.pi/90.)/wavelength
    if sthovl_min <= 0:
        sthovl_min = 0.0001
    sthovl_max = numpy.sin(0.5*ttheta_zs.max() + numpy.pi/90.)/wavelength
    if sthovl_max <= sthovl_min:
        sthovl_max = sthovl_min+0.01
        if sthovl_max >= 1.:
            sthovl_max =  0.99999/wavelength
        
    background_ttheta = dict_pd["background_ttheta"]
    background_intensity = dict_pd["background_intensity"]
    flags_background_intensity = dict_pd["flags_background_intensity"]

    flag_background_intensity = numpy.any(flags_background_intensity)
    if (flag_use_precalculated_data and ("signal_background" in dict_in_out) and not(flag_background_intensity)):
        signal_background = dict_in_out["signal_background"]
    else:
        signal_background, dder_s_bkgr = calc_background(ttheta, background_ttheta, background_intensity,
            flag_background_intensity= (flag_background_intensity and flag_calc_analytical_derivatives))
        dict_in_out["signal_background"] = signal_background

    pd_phase_name = dict_pd["phase_name"]
    pd_phase_scale = dict_pd["phase_scale"]
    pd_phase_resolution_parameters = dict_pd["phase_resolution_parameters"] # U_phase, V_phase, W_phase, X_phase, Y_phase
    pd_phase_ig = dict_pd["phase_ig"] # IG_phase
    flags_pd_phase_scale = dict_pd["flags_phase_scale"]
    flags_pd_phase_resolution_parameters = dict_pd["flags_phase_resolution_parameters"] # U_phase, V_phase, W_phase, X_phase, Y_phase
    flags_pd_phase_ig = dict_pd["flags_phase_ig"] # IG_phase

    resolution_parameters = dict_pd["resolution_parameters"] # U, V, W, X, Y
    if "asymmetry_parameters" in dict_pd_keys:
        asymmetry_parameters = dict_pd["asymmetry_parameters"] # p1, p2, p3, p4
        flags_asymmetry_parameters = dict_pd["flags_asymmetry_parameters"] 
        flag_asymmetry_parameters = numpy.any(flags_asymmetry_parameters)
        p_1, p_2, p_3, p_4 = asymmetry_parameters[0], asymmetry_parameters[1], asymmetry_parameters[2], asymmetry_parameters[3]

    else:
        p_1, p_2, p_3, p_4 = 0., 0., 0., 0.
        flag_asymmetry_parameters = False

    flags_resolution_parameters = dict_pd["flags_resolution_parameters"] 
    
    if "texture_name" in dict_pd_keys:
        flag_texture = True
        pd_texture_name = dict_pd["texture_name"]
        pd_texture_g1 = dict_pd["texture_g1"]
        pd_texture_g2 = dict_pd["texture_g2"]
        pd_texture_axis = dict_pd["texture_axis"]
        pd_flags_texture_g1 = dict_pd["flags_texture_g1"]
        pd_flags_texture_g2 = dict_pd["flags_texture_g2"]
        pd_flags_texture_axis = dict_pd["flags_texture_axis"]
    else:
        flag_texture = False
    
    k = dict_pd["k"]
    cthm = dict_pd["cthm"]
    
    lorentz_factor, dder_lf = calc_lorentz_factor(ttheta_zs, k=k, cthm=cthm, flag_ttheta=flags_offset_ttheta)
    dict_in_out["lorentz_factor"] = lorentz_factor


    total_signal_plus = numpy.zeros_like(ttheta_zs)
    total_signal_minus = numpy.zeros_like(ttheta_zs)
    for p_name, p_scale, p_resolution, p_ig, flags_p_scale, flags_p_resolution, flags_p_ig in zip(pd_phase_name, 
            pd_phase_scale, pd_phase_resolution_parameters.transpose(), pd_phase_ig,
            flags_pd_phase_scale, flags_pd_phase_resolution_parameters.transpose(), flags_pd_phase_ig):
        p_name = p_name.lower()
        flag_phase_texture = False
        if flag_texture:
            ind_texture = numpy.argwhere(pd_texture_name==p_name)
            if ind_texture.shape[0] != 0:
                texture_g1 = pd_texture_g1[ind_texture[0]]
                texture_g2 = pd_texture_g2[ind_texture[0]]
                texture_axis = pd_texture_axis[:, ind_texture[0]]
                flag_phase_texture = True
                flags_texture_g1 = pd_flags_texture_g1[ind_texture[0]]
                flags_texture_g2 = pd_flags_texture_g2[ind_texture[0]]
                flags_texture_axis = pd_flags_texture_axis[:, ind_texture[0]]

        ind_phase = phase_name.index(p_name)
        dict_crystal = dict_crystals[ind_phase]
        dict_in_out_keys = dict_in_out.keys()
        if f"dict_in_out_{p_name:}" in dict_in_out_keys:
            dict_in_out_phase = dict_in_out[f"dict_in_out_{p_name:}"]
        else:
            dict_in_out_phase = {}
            dict_in_out[f"dict_in_out_{p_name:}"] = dict_in_out_phase

        dict_in_out_phase_keys = dict_in_out_phase.keys()
        dict_crystal_keys = dict_crystal.keys()

        if "reduced_symm_elems" in dict_crystal_keys:
            reduced_symm_elems  = dict_crystal["reduced_symm_elems"]
            translation_elems = dict_crystal["translation_elems"]
        elif "full_symm_elems" in dict_crystal_keys:
            full_symm_elems = dict_crystal["full_symm_elems"]
        elif "full_mcif_elems" in dict_crystal_keys:
            full_mcif_elems  = dict_crystal["full_mcif_elems"]

        unit_cell_parameters = dict_crystal["unit_cell_parameters"]
        flags_unit_cell_parameters = dict_crystal["flags_unit_cell_parameters"]
        flag_unit_cell_parameters = numpy.any(flags_unit_cell_parameters)

        if flag_unit_cell_parameters:
            sc_uc = dict_crystal["sc_uc"]
            v_uc = dict_crystal["v_uc"]
            unit_cell_parameters = numpy.dot(sc_uc, unit_cell_parameters) + v_uc


        if (flag_use_precalculated_data and 
                ("index_hkl" in dict_in_out_phase_keys) and 
                ("multiplicity_hkl" in dict_in_out_phase_keys) and not(flag_unit_cell_parameters or flags_offset_ttheta)):
            index_hkl = dict_in_out_phase["index_hkl"]
            multiplicity_hkl = dict_in_out_phase["multiplicity_hkl"]
        else:
            if flag_phase_texture:
                reduced_symm_elems_p1 = numpy.array([[0], [0], [0], [1], [1], [0], [0], [0], [1], [0], [0], [0], [1]], dtype=int)
                translation_elems_p1 = numpy.array([[0], [0], [0], [1]], dtype=int)
                index_hkl, multiplicity_hkl = calc_index_hkl_multiplicity_in_range(
                    sthovl_min, sthovl_max, unit_cell_parameters, reduced_symm_elems_p1, translation_elems_p1, False)
            else:
                if "reduced_symm_elems" in dict_crystal_keys:
                    centrosymmetry = dict_crystal["centrosymmetry"]
                    index_hkl, multiplicity_hkl = calc_index_hkl_multiplicity_in_range(
                        sthovl_min, sthovl_max, unit_cell_parameters, reduced_symm_elems, translation_elems, centrosymmetry)
                else:
                    translation_elems_p1 = numpy.array([[0], [0], [0], [1]], dtype=int)
                    index_hkl, multiplicity_hkl = calc_index_hkl_multiplicity_in_range(
                        sthovl_min, sthovl_max, unit_cell_parameters, full_mcif_elems[:13], translation_elems_p1, False)

            if (("index_hkl" in dict_in_out_phase_keys) and flag_use_precalculated_data):
                if index_hkl.shape != dict_in_out_phase["index_hkl"].shape:
                    flag_use_precalculated_data = False
                else: 
                    flag_use_precalculated_data = numpy.all(numpy.logical_and(dict_in_out_phase["index_hkl"], index_hkl))

            dict_in_out_phase["index_hkl"] = index_hkl
            dict_in_out_phase["multiplicity_hkl"] = multiplicity_hkl

        flag_sthovl_hkl = flag_unit_cell_parameters
        sthovl_hkl, dder_sthovl_hkl = calc_sthovl_by_unit_cell_parameters(index_hkl,
            unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)

        flag_ttheta_hkl = flag_sthovl_hkl or flags_wavelength
        ttheta_hkl = 2*numpy.arcsin(sthovl_hkl*wavelength)
        dict_in_out_phase["ttheta_hkl"] = ttheta_hkl + offset_ttheta
        if radiation[0].startswith("neutrons"):
            f_nucl, dder_f_nucl = calc_f_nucl_by_dictionary(
                dict_crystal, dict_in_out_phase, flag_use_precalculated_data=flag_use_precalculated_data)
            flag_f_nucl = len(dder_f_nucl.keys()) > 0

            flag_para = False
            if (("atom_para_index" in dict_crystal_keys) and ("atom_para_susceptibility" in dict_crystal_keys)):
                sft_ccs, dder_sft_ccs = calc_sft_ccs_by_dictionary(
                    dict_crystal, dict_in_out_phase, flag_use_precalculated_data=flag_use_precalculated_data)
                flag_sft_ccs  = len(dder_sft_ccs.keys()) > 0

                flag_matrix_t = flag_unit_cell_parameters
                matrix_t, dder_matrix_t = calc_matrix_t(
                    index_hkl, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)

                flag_tensor_sigma = flag_sft_ccs or flag_unit_cell_parameters
                tensor_sigma, dder_tensor_sigma = calc_m1_m2_m1t(matrix_t, sft_ccs, flag_m1=flag_sft_ccs, flag_m2=flag_unit_cell_parameters)
                flag_para = True

            flag_ordered = False
            if "atom_ordered_index" in dict_crystal_keys:
                f_m_perp_o_ccs, dder_f_m_perp_o_ccs = calc_f_m_perp_ordered_by_dictionary(
                    dict_crystal, dict_in_out_phase, flag_use_precalculated_data=flag_use_precalculated_data)
                flag_f_m_perp_o =  len(dder_f_m_perp_o_ccs.keys()) > 0
                flag_ordered = True

                flag_matrix_t = flag_unit_cell_parameters
                matrix_t, dder_matrix_t = calc_matrix_t(
                    index_hkl, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)
                f_m_perp_o, dder_f_m_perp_o = calc_m_v(matrix_t, f_m_perp_o_ccs, flag_m=flag_unit_cell_parameters, flag_v=flag_f_m_perp_o)

            if flag_para and not(flag_ordered):
                flag_iint_plus_minus = flag_f_nucl or flag_tensor_sigma or flags_beam_polarization or flags_flipper_efficiency
                if (("iint_plus" in dict_in_out_phase_keys) and ("iint_minu" in dict_in_out_phase_keys) and
                        flag_use_precalculated_data and not(flag_iint_plus_minus)):
                    iint_plus, iint_minus = dict_in_out_phase["iint_plus"], dict_in_out_phase["iint_minus"]
                else:
                    iint_plus, iint_minus, dder_plus, dder_minus = calc_powder_iint_1d_para(
                        f_nucl, tensor_sigma, beam_polarization, flipper_efficiency, magnetic_field,
                        flag_f_nucl=flag_f_nucl, flag_tensor_sigma=flag_tensor_sigma,
                        flag_polarization=flags_beam_polarization, flag_flipper=flags_flipper_efficiency)
            elif not(flag_para) and flag_ordered:
                flag_iint_plus_minus = flag_f_nucl or flag_f_m_perp_o or flags_beam_polarization or flags_flipper_efficiency
                if (("iint_plus" in dict_in_out_phase_keys) and ("iint_minus" in dict_in_out_phase_keys) and
                        flag_use_precalculated_data and not(flag_iint_plus_minus)):
                    iint_plus, iint_minus = dict_in_out_phase["iint_plus"], dict_in_out_phase["iint_minus"]
                else:
                    iint_plus, dder_plus = calc_powder_iint_1d_ordered(
                        f_nucl, f_m_perp_o,
                        flag_f_nucl=flag_f_nucl and flag_calc_analytical_derivatives,
                        flag_f_m_perp=flag_f_m_perp_o and flag_calc_analytical_derivatives)
                    iint_minus = iint_plus
                    dder_minus = dder_plus
            elif flag_para and flag_ordered:
                flag_iint_plus_minus = flag_f_nucl or flag_tensor_sigma or flag_f_m_perp_o or flags_beam_polarization or flags_flipper_efficiency
                if (("iint_plus" in dict_in_out_phase_keys) and ("iint_minu" in dict_in_out_phase_keys) and
                        flag_use_precalculated_data and not(flag_iint_plus_minus)):
                    iint_plus, iint_minus = dict_in_out_phase["iint_plus"], dict_in_out_phase["iint_minus"]
                else:
                    iint_plus, iint_minus, dder_plus, dder_minus = calc_powder_iint_1d_mix(
                        f_nucl, f_m_perp_o, tensor_sigma, beam_polarization, flipper_efficiency, magnetic_field,
                        flag_f_nucl=flag_f_nucl and flag_calc_analytical_derivatives,
                        flag_f_m_perp_ordered=flag_f_m_perp_o and flag_calc_analytical_derivatives,
                        flag_tensor_sigma=flag_tensor_sigma and flag_calc_analytical_derivatives,
                        flag_polarization=flags_beam_polarization and flag_calc_analytical_derivatives,
                        flag_flipper=flags_flipper_efficiency and flag_calc_analytical_derivatives)
            else:
                iint_plus = numpy.square(numpy.abs(f_nucl))
                iint_minus = numpy.square(numpy.abs(f_nucl))

            dict_in_out_phase["iint_plus"] = iint_plus
            dict_in_out_phase["iint_minus"] = iint_minus
        elif radiation[0].startswith("X-rays"):
            f_charge, dder_f_charge = calc_f_charge_by_dictionary(
                dict_crystal, wavelength, dict_in_out_phase, flag_use_precalculated_data=flag_use_precalculated_data)
            flag_f_charge = len(dder_f_charge.keys()) > 0

            iint = numpy.square(numpy.abs(f_charge))
            # FIXME: preparation for XMD
            iint_plus = iint 
            iint_minus = iint

            dict_in_out_phase["iint_plus"] = iint_plus
            dict_in_out_phase["iint_minus"] = iint_minus

        if flag_phase_texture:
            flag_texture_g1 = numpy.any(flags_texture_g1)
            flag_texture_g2 = numpy.any(flags_texture_g2)
            flag_texture_axis = numpy.any(flags_texture_axis)
            flag_hh = numpy.any([flag_texture_g1, flag_texture_g2, flag_texture_axis])
            if (flag_use_precalculated_data and 
                    ("preferred_orientation" in dict_in_out_phase_keys) and
                    not(flag_hh)):
                preferred_orientation = dict_in_out_phase["preferred_orientation"]
            else:
                preferred_orientation, dder_po = calc_preferred_orientation_pd(
                    index_hkl, texture_g1, texture_g2, texture_axis, unit_cell_parameters, 
                    flag_texture_g1=flag_texture_g1 and flag_calc_analytical_derivatives,
                    flag_texture_g2=flag_texture_g2 and flag_calc_analytical_derivatives,
                    flag_texture_axis=flag_texture_axis and flag_calc_analytical_derivatives)
                dict_in_out_phase["preferred_orientation"] = preferred_orientation
        
        flag_rp = numpy.any(flags_p_resolution) or numpy.any(flags_resolution_parameters)
        
        hh = resolution_parameters + p_resolution
        u, v, w, x, y = hh[0], hh[1], hh[2], hh[3], hh[4]
        
        profile_pv, dder_pv = calc_profile_pseudo_voight(ttheta_zs, ttheta_hkl, u, v, w, p_ig, x, y,
            p_1, p_2, p_3, p_4, 
            flag_ttheta=flags_offset_ttheta,
            flag_ttheta_hkl=flag_ttheta_hkl, flag_u=flag_rp,
            flag_v=flag_rp, flag_w=flag_rp, flag_i_g=flags_p_ig,
            flag_x=flag_rp, flag_y=flag_rp, 
            flag_p_1=flag_asymmetry_parameters, flag_p_2=flag_asymmetry_parameters,
            flag_p_3=flag_asymmetry_parameters, flag_p_4=flag_asymmetry_parameters)
        dict_in_out_phase["profile_pv"] = profile_pv


        # flags_p_scale
        iint_m_plus = iint_plus * multiplicity_hkl
        iint_m_minus = iint_minus * multiplicity_hkl
        lf = calc_lorentz_factor(ttheta_hkl, k=k, cthm=cthm, flag_ttheta=None)[0]
        dict_in_out_phase["iint_plus_with_factors"] = 0.5 * p_scale * lf * iint_m_plus
        dict_in_out_phase["iint_minus_with_factors"] = 0.5 * p_scale * lf * iint_m_minus
        if flag_texture:
            # 0.5 to have the same meaning for the scale factor as in FullProf
            signal_plus = 0.5 * p_scale * lorentz_factor * (profile_pv * (iint_m_plus * preferred_orientation)[na, :]).sum(axis=1) # sum over hkl
            signal_minus = 0.5 * p_scale * lorentz_factor * (profile_pv * (iint_m_minus * preferred_orientation)[na, :]).sum(axis=1) 
            dict_in_out_phase["iint_plus_with_factors"] *= preferred_orientation
            dict_in_out_phase["iint_minus_with_factors"] *= preferred_orientation
        else:
            signal_plus = 0.5 * p_scale * lorentz_factor * (profile_pv * iint_m_plus[na, :]).sum(axis=1) 
            signal_minus = 0.5 * p_scale * lorentz_factor * (profile_pv * iint_m_minus[na, :]).sum(axis=1) 
        
        dict_in_out_phase["signal_plus"] = signal_plus
        dict_in_out_phase["signal_minus"] = signal_minus
        total_signal_plus += signal_plus
        total_signal_minus += signal_minus

    if flag_dict:
        dict_in_out["signal_plus"] = total_signal_plus
        dict_in_out["signal_minus"] = total_signal_minus

    if ("signal_exp_plus" in dict_pd_keys) and ("signal_exp_minus" in dict_pd_keys):
        signal_exp_plus = dict_pd["signal_exp_plus"]
        signal_exp_minus = dict_pd["signal_exp_minus"]
        if flag_dict:
            dict_in_out["signal_exp_plus"] = signal_exp_plus 
            dict_in_out["signal_exp_minus"] = signal_exp_minus 
        flag_chi_sq_sum, flag_chi_sq_difference = True, True

        if "flag_chi_sq_sum" in dict_pd_keys:
            flag_chi_sq_sum = dict_pd["flag_chi_sq_sum"]

        if "flag_chi_sq_difference" in dict_pd_keys:
            flag_chi_sq_difference = dict_pd["flag_chi_sq_difference"]

        if flag_chi_sq_sum:
            signal_exp = signal_exp_plus[0, :] + signal_exp_minus[0, :]
            signal_sigma = numpy.sqrt(numpy.square(signal_exp_plus[1, :]) + numpy.square(signal_exp_minus[1, :]))

    else:
        signal_exp = dict_pd["signal_exp"][0,:]
        signal_sigma = dict_pd["signal_exp"][1,:]
        if flag_dict:
            dict_in_out["signal_exp"] = dict_pd["signal_exp"] 
        flag_chi_sq_sum = True
        flag_chi_sq_difference = False
        

    chi_sq = 0.
    n_point = 0
    if flag_chi_sq_sum:
        in_points = numpy.logical_not(excluded_points)
        total_signal_sum = total_signal_plus + total_signal_minus + signal_background
        chi_sq_sum = ((numpy.square((signal_exp - total_signal_sum)/signal_sigma)*in_points)).sum(axis=0)
        chi_sq += chi_sq_sum
        n_point += numpy.sum(in_points)
    
    if flag_chi_sq_difference:
        signal_exp_diff = signal_exp_plus[0, :] - signal_exp_minus[0, :]
        signal_sigma_diff = numpy.sqrt(numpy.square(signal_exp_plus[1, :]) + numpy.square(signal_exp_minus[1, :]))
        total_signal_diff = total_signal_plus - total_signal_minus
        chi_sq_diff = (numpy.square((signal_exp_diff - total_signal_diff)/signal_sigma_diff)).sum(axis=0)
        chi_sq += chi_sq_diff
        n_point += signal_exp_diff.shape[0]
    if numpy.isnan(chi_sq):
        chi_sq = 1e30


    flags_pd = get_flags(dict_pd)
    l_flags_crystal = [get_flags(dict_crystal) for dict_crystal in dict_crystals]

    l_parameter_name = []
    for way, flags in flags_pd.items():
        pd_type_name = dict_pd["type_name"]
        ind_1d = numpy.atleast_1d(numpy.argwhere(flags)) #.flatten()
        parameter_name = [(pd_type_name, ) + way + (tuple(ind_1d[ind,:]), ) for ind in range(ind_1d.shape[0])]
        l_parameter_name.extend(parameter_name)

    for flags_crystal, dict_crystal in zip(l_flags_crystal, dict_crystals):
        for way, flags in flags_crystal.items():
            crystal_type_name = dict_crystal["type_name"]
            ind_1d = numpy.atleast_1d(numpy.argwhere(flags)) #.flatten()
            parameter_name = [(crystal_type_name, ) + way + (tuple(ind_1d[ind,:]), ) for ind in range(ind_1d.shape[0])]
            l_parameter_name.extend(parameter_name)
    
    if flag_background_intensity:
        pass

    der_chi_sq = numpy.zeros((len(l_parameter_name), ), dtype=float) 
    dder_chi_sq = numpy.zeros((len(l_parameter_name), len(l_parameter_name)), dtype=float)

    return chi_sq, n_point, der_chi_sq, dder_chi_sq, l_parameter_name
    